#############################################################################
##
#W  blockobj.gd          GAP 4 package CTBlocks                 Thomas Breuer
##
#Y  Copyright (C)  2013,   Lehrstuhl D für Mathematik,  RWTH Aachen,  Germany
##


#############################################################################
##
##  <#GAPDoc Label="intro:blockobj">
##  This section introduces &GAP; objects that represent <M>p</M>-blocks of
##  character tables.
##  One may argue that such objects are not necessary because one can simply
##  replace the object that represents the <M>b</M>-th <M>p</M>-block of the
##  character table <M>t</M> by the triple <M>[ t, p, b ]</M>.
##  Another argument against these objects is that the documentation of
##  functions for them is longer than the &GAP; code that implements them.
##  On the other hand, such objects are convenient as inputs and outputs of
##  operations, and for caching results of computations as values of
##  attributes, such as <Ref Attr="Irr" Label="for a block object"/>
##  and <Ref Attr="IBr" Label="for a block object"/>.
##  <P/>
##  Thus we offer such &GAP; objects, but we offer also variants of the
##  relevant functions that accept the arguments <M>t</M>, <M>p</M>, <M>b</M>
##  instead.
##  <P/>
##  See Section <Ref Subsect="IsBlockOfCharacterTable"/> for technical
##  details.
##  <#/GAPDoc>
##


CTBlocks:= rec();


#############################################################################
##
#C  IsBlockOfCharacterTable
##
##  <#GAPDoc Label="IsBlockOfCharacterTable">
##  <ManSection>
##  <Heading>Technicalities of block objects for character tables
##  </Heading>
##  <Filt Name="IsBlockOfCharacterTable" Arg="B" Type="category"/>
##
##  <Description>
##  A <M>p</M>-block of a character table can be represented by a &GAP; object
##  in the filter <Ref Func="IsBlockOfCharacterTable"/>.
##  The idea is that such an object is created by <Ref Oper="Block"/> from
##  the defining data, and that there are various operations that can
##  take a block object as arguments (for example
##  attributes such as <Ref Attr="Defect" Label="for a block object"/>
##  and <Ref Attr="Irr" Label="for a block object"/>)
##  or can return block objects (for example <Ref Oper="PrincipalBlock"/>).
##  <P/>
##  Block objects aren't domains in the sense of
##  <Ref Chap="Domains" BookName="ref"/>,
##  in particular they do not have elements;
##  one can use <Ref Attr="Irr" Label="for a block object"/>
##  and <Ref Attr="IBr" Label="for a block object"/> for accessing the
##  ordinary or modular irreducible characters that <Q>belong to</Q> the
##  given block.
##  <P/>
##  <#Include Label="compare_blocks">
##  <P/>
##  The declarations of operations and the installations of methods for these
##  block objects should not cause conflicts to situations where blocks of
##  algebras, in the sense of direct summands, are considered.
##  (At the moment, there seems to be no &GAP; package with such
##  functionality.)
##  <P/>
##  <Example><![CDATA[
##  gap> tbl:= CharacterTable( "D8" );;
##  gap> tbl2:= CharacterTable( "A4" );;
##  gap> IsBlockOfCharacterTable( tbl );
##  false
##  gap> IsBlockOfCharacterTable( Block( tbl, 2, 1 ) );
##  true
##  gap> Block( tbl, 2, 1 ) = Block( tbl2, 2, 1 );
##  false
##  gap> Block( tbl, 2, 1 ) < Block( tbl2, 2, 1 );
##  true
##  ]]></Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareCategory( "IsBlockOfCharacterTable", IsObject );


#############################################################################
##
#O  Block( <tbl>, <p>, <b> )
##
##  <#GAPDoc Label="Block">
##  <ManSection>
##  <Oper Name="Block" Arg="tbl, p, b"/>
##
##  <Description>
##  For an ordinary character table <A>tbl</A>, a prime integer <A>p</A>,
##  and a positive integer <A>b</A>,
##  <Ref Oper="Block"/> returns the block object that represents the
##  <A>b</A>-th <A>p</A>-block of <A>tbl</A> if this block exists,
##  otherwise <K>fail</K> is returned.
##  <P/>
##  <Example>
##  gap> Block( CharacterTable( "A5" ), 2, 1 );
##  Block( CharacterTable( "A5" ), 2, 1 )
##  gap> Block( CharacterTable( "A5" ), 2, 3 );
##  fail
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareOperation( "Block", [ IsOrdinaryTable, IsPosInt, IsPosInt ] );


#############################################################################
##
#A  UnderlyingCharacterTable( <B> )
#A  UnderlyingCharacteristic( <B> )
#A  NumberOfBlock( <B> )
##
##  <#GAPDoc Label="subsect:blocks_defining">
##  <ManSection>
##  <Heading>Defining attributes of block objects</Heading>
##  <Attr Name="UnderlyingCharacterTable" Arg="B"
##   Label="for a block object"/>
##  <Attr Name="UnderlyingCharacteristic" Arg="B"
##   Label="for a block object"/>
##  <Attr Name="NumberOfBlock" Arg="B" Label="for a block object"/>
##
##  <Description>
##  These are the defining attributes of the block object <A>B</A>.
##  There are no methods for computing their values,
##  the values must be set upon creation of the block object.
##  <P/>
##  For the <M>b</M>-th <M>p</M>-block of the character table <M>t</M>,
##  the values of
##  <Ref Attr="UnderlyingCharacterTable" Label="for a block object"/>,
##  <Ref Attr="UnderlyingCharacteristic" Label="for a block object"/>,
##  and <Ref Attr="NumberOfBlock" Label="for a block object"/>
##  are <M>t</M>, <M>p</M>, and <M>b</M>, respectively.
##  <P/>
##  <Example>
##  gap> b:= Block( CharacterTable( "A5" ), 2, 1 );
##  Block( CharacterTable( "A5" ), 2, 1 )
##  gap> UnderlyingCharacterTable( b );
##  CharacterTable( "A5" )
##  gap> UnderlyingCharacteristic( b );
##  2
##  gap> NumberOfBlock( b );
##  1
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareAttribute( "UnderlyingCharacterTable", IsBlockOfCharacterTable );

DeclareAttribute( "UnderlyingCharacteristic", IsBlockOfCharacterTable );

DeclareAttribute( "NumberOfBlock", IsBlockOfCharacterTable );


#############################################################################
##
#F  PBlocks( <tbl>, <p> )
##
##  <#GAPDoc Label="PBlocks">
##  <ManSection>
##  <Func Name="PBlocks" Arg="tbl, p"/>
##
##  <Description>
##  For an ordinary character table <A>tbl</A> and a prime integer <A>p</A>,
##  <Ref Func="PBlocks"/> returns the list of block objects that represent
##  the <A>p</A>-blocks of <A>tbl</A>.
##  <P/>
##  <Example>
##  gap> PBlocks( CharacterTable( "A5" ), 2 );
##  [ Block( CharacterTable( "A5" ), 2, 1 ), 
##    Block( CharacterTable( "A5" ), 2, 2 ) ]
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareAttribute( "ComputedPBlocks",
    IsOrdinaryTable, "mutable" );

DeclareOperation( "PBlocksOp", [ IsOrdinaryTable, IsPosInt ] );

DeclareGlobalFunction( "PBlocks" );


#############################################################################
##
#F  PRegularTable( <tbl>, <p> )
#A  ComputedPRegularTables( <tbl>, <p> )
##
##  <#GAPDoc Label="PRegularTable">
##  <ManSection>
##  <Func Name="PRegularTable" Arg="tbl, p"/>
##  <Attr Name="ComputedPRegularTables" Arg="tbl"/>
##
##  <Description>
##  For an ordinary character table <A>tbl</A> and a prime integer <A>p</A>,
##  <Ref Func="PRegularTable"/> returns the same as
##  <Ref Func="CharacterTableRegular" BookName="ref"/>.
##  The only difference is that the results of <Ref Func="PRegularTable"/>
##  are cached in <A>tbl</A>,
##  via the attribute <Ref Attr="ComputedPRegularTables"/>.
##  <P/>
##  The purpose of this attribute is to store <M>p</M>-regular character
##  tables (see <Ref Func="CharacterTableRegular" BookName="ref"/>
##  for those primes <M>p</M> for which the irreducible characters of the
##  <M>p</M>-modular character table are not available.
##  It may still be possible to provide the irreducibles of certain blocks,
##  and for that, it is necessary to assign these characters to a suitable
##  modular character table.
##  <P/>
##  The cached tables are used for example by
##  <Ref Attr="IBr" Label="for a block object"/>.
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareGlobalFunction( "PRegularTable" );

DeclareAttribute( "ComputedPRegularTables",
    IsOrdinaryTable, "mutable" );


#############################################################################
##
#O  PrincipalBlock( <tbl>, <p> )
##
##  <#GAPDoc Label="PrincipalBlock">
##  <ManSection>
##  <Oper Name="PrincipalBlock" Arg="tbl, p"/>
##
##  <Description>
##  For an ordinary character table <A>tbl</A> and a prime integer
##  <Ref Oper="PrincipalBlock"/> returns the block object for the
##  <A>p</A>-block of <A>tbl</A> that contains the trivial character
##  of <A>tbl</A>.
##  <P/>
##  Note that the trivial character need not be the first character in the
##  list of irreducibles of <A>tbl</A>,
##  so the principal block need not have number <M>1</M>.
##  <P/>
##  <Example>
##  gap> PrincipalBlock( CharacterTable( "A5" ), 2 );
##  Block( CharacterTable( "A5" ), 2, 1 )
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareOperation( "PrincipalBlock", [ IsOrdinaryTable, IsPosInt ] );


#############################################################################
##
#A  Irr( <B> )
#O  Irr( <tbl>, <p>, <b> )
##
##  <#GAPDoc Label="Irr">
##  <ManSection>
##  <Attr Name="Irr" Arg="B"
##   Label="for a block object"/>
##  <Oper Name="Irr" Arg="ordtbl, p, b"
##   Label="for character table, characteristic, and block number"/>
##
##  <Description>
##  The arguments can be either a block object <A>B</A> which represents the
##  <A>b</A>-th <A>p</A>-block of the ordinary character table <A>ordtbl</A>,
##  or the three values <A>ordtbl</A>, <A>p</A>, and <A>b</A>.
##  <P/>
##  <Ref Oper="Irr" Label="for a block object"/> returns the sublist of those
##  irreducible characters of <A>ordtbl</A> that belong to the block,
##  in the same ordering as in the <Ref Oper="Irr" BookName="ref"/> list
##  of <A>ordtbl</A>.
##  <P/>
##  <Example>
##  gap> irr:= Irr( Block( CharacterTable( "A5" ), 2, 1 ) );
##  [ Character( CharacterTable( "A5" ), [ 1, 1, 1, 1, 1 ] ), 
##    Character( CharacterTable( "A5" ),
##    [ 3, -1, 0, -E(5)-E(5)^4, -E(5)^2-E(5)^3 ] ), 
##    Character( CharacterTable( "A5" ),
##    [ 3, -1, 0, -E(5)^2-E(5)^3, -E(5)-E(5)^4 ] ), 
##    Character( CharacterTable( "A5" ), [ 5, 1, -1, 0, 0 ] ) ]
##  gap> irr = Irr( CharacterTable( "A5" ), 2, 1 );
##  true
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareAttribute( "Irr", IsBlockOfCharacterTable );

DeclareOperation( "Irr", [ IsOrdinaryTable, IsPosInt, IsPosInt ] );


#############################################################################
##
#A  IBr( <B> )
#O  IBr( <modtbl>, <b> )
##
##  <#GAPDoc Label="IBr">
##  <ManSection>
##  <Attr Name="IBr" Arg="B"
##   Label="for a block object"/>
##  <Oper Name="IBr" Arg="modtbl, b"
##   Label="for Brauer table and block number"/>
##
##  <Description>
##  The arguments can be either a block object <A>B</A> which represents the
##  <A>b</A>-th <M>p</M>-block of the ordinary character table whose
##  <M>p</M>-modular Brauer table is <A>modtbl</A>,
##  or the two values <A>modtbl</A> and <A>b</A>.
##  <P/>
##  <Ref Oper="IBr" Label="for a block object"/> returns the sublist of those
##  irreducible characters of <A>modtbl</A> that belong to the block,
##  in the same ordering as in the <Ref Oper="Irr" BookName="ref"/> list
##  of <A>modtbl</A>.
##  <P/>
##  If <A>B</A> is given, it may happen that the irreducible <M>p</M>-modular
##  Brauer characters for the character table that defines <A>B</A> are not
##  available.
##  In this case,
##  <Ref Oper="IBr" Label="for a block object"/> returns <K>fail</K>.
##  <P/>
##  <Example>
##  gap> ibr:= IBr( Block( CharacterTable( "A5" ), 2, 1 ) );
##  [ Character( BrauerTable( "A5", 2 ), [ 1, 1, 1, 1 ] ), 
##    Character( BrauerTable( "A5", 2 ),
##    [ 2, -1, E(5)+E(5)^4, E(5)^2+E(5)^3 ] ), 
##    Character( BrauerTable( "A5", 2 ),
##    [ 2, -1, E(5)^2+E(5)^3, E(5)+E(5)^4 ] ) ]
##  gap> ibr = IBr( CharacterTable( "A5" ) mod 2, 1 );
##  true
##  gap> IBr( Block( CharacterTable( "M" ), 2, 1 ) );
##  fail
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareAttribute( "IBr", IsBlockOfCharacterTable );

DeclareOperation( "IBr", [ IsBrauerTable, IsPosInt ] );


#############################################################################
##
#A  DecompositionMatrix( <B> )
##
##  <#GAPDoc Label="DecompositionMatrix">
##  <ManSection>
##  <Attr Name="DecompositionMatrix" Arg="B"
##   Label="for a block object"/>
##
##  <Description>
##  For a <M>p</M>-block <A>B</A> of a character table <M>t</M>, say.
##  <Ref Attr="DecompositionMatrix" Label="for a block object"/> returns the
##  decomposition matrix of <A>B</A>,
##  that is, the matrix whose rows and columns are indexed by the
##  ordinary irreducible characters and the <M>p</M>-modular Brauer
##  characters, respectively, that belong to <A>B</A>.
##  <P/>
##  If the <M>p</M>-modular character table of <M>t</M> is not known then
##  <K>fail</K> is returned.
##  <P/>
##  A variant without block objects is
##  <Ref Oper="DecompositionMatrix" BookName="ref"/>.
##  <P/>
##  <Example>
##  gap> DecompositionMatrix( Block( CharacterTable( "A5" ), 2, 1 ) );
##  [ [ 1, 0, 0 ], [ 1, 0, 1 ], [ 1, 1, 0 ], [ 1, 1, 1 ] ]
##  gap> DecompositionMatrix( Block( CharacterTable( "A5" ), 2, 2 ) );
##  [ [ 1 ] ]
##  gap> DecompositionMatrix( Block( CharacterTable( "M" ), 2, 1 ) );
##  fail
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareAttribute( "DecompositionMatrix", IsBlockOfCharacterTable );


#############################################################################
##
#A  DimensionsOfDecompositionMatrix( <B> )
##
##  <#GAPDoc Label="DimensionsOfDecompositionMatrix">
##  <ManSection>
##  <Attr Name="DimensionsOfDecompositionMatrix" Arg="B"
##   Label="for a block object"/>
##
##  <Description>
##  For a <M>p</M>-block <A>B</A> of a character table <M>t</M>, say.
##  <Ref Attr="DimensionsOfDecompositionMatrix" Label="for a block object"/>
##  returns the list
##  <M>[ k(</M><A>B</A><M>), l(</M><A>B</A><M>) ]</M> of the numbers of rows
##  and columns of the decomposition matrix of <A>B</A>.
##  That is, <M>k(</M><A>B</A><M>)</M> is the number of ordinary irreducible
##  characters in <A>B</A>,
##  and <M>l(</M><A>B</A><M>)</M> is the number of irreducible Brauer
##  characters in <A>B</A>.
##  <P/>
##  Note that these numbers can be computed from the ordinary character table
##  <M>t</M>, whereas
##  <Ref Attr="DecompositionMatrix" Label="for a block object"/>
##  requires the irreducible <M>p</M>-modular Brauer characters of <M>t</M>.
##  <P/>
##  <Example>
##  gap> b:= Block( CharacterTable( "M" ), 2, 1 );;
##  gap> DimensionsOfDecompositionMatrix( b );
##  [ 183, 55 ]
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareAttribute( "DimensionsOfDecompositionMatrix",
    IsBlockOfCharacterTable );


#############################################################################
##
#A  Dimension( <B> )
##
##  <#GAPDoc Label="Dimension">
##  <ManSection>
##  <Attr Name="Dimension" Arg="B" Label="for a block object"/>
##
##  <Description>
##  The dimension of the <M>p</M>-block <A>B</A>
##  is <M>\sum_{{\chi}} \chi(1)^2</M>,
##  where the summation runs over the ordinary irreducible characters
##  <M>\chi</M> of <A>B</A>.
##  <P/>
##  This is the dimension of the algebra <M>e FG</M>,
##  where <M>e</M> is the block idempotent of <A>B</A>
##  (see <Ref Attr="CoefficientsOfOsimaIdempotent"
##  Label="for a block object"/>)
##  and <M>G</M> is the underlying group of <A>B</A>.
##  <P/>
##  (For details, see <Cite Key="Nav98" Where="Thm. 3.14"/>.)
##  <P/>
##  <Example>
##  gap> List( PBlocks( CharacterTable( "2.A5" ), 3 ), Dimension );
##  [ 42, 9, 9, 24, 36 ]
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareAttribute( "Dimension", IsBlockOfCharacterTable );


#############################################################################
##
#F  FrobeniusCharacterValueExt( <value>, <p> )
##
##  <#GAPDoc Label="FrobeniusCharacterValueExt">
##  <ManSection>
##  <Func Name="FrobeniusCharacterValueExt" Arg="value, p"/>
##
##  <Description>
##  The &GAP; library function
##  <Ref Func="FrobeniusCharacterValue" BookName="ref"/> works only for those
##  cyclotomics <A>value</A> whose conductor is coprime to <A>p</A>.
##  In order to implement the ring homomorphism from all cyclotomics whose
##  coefficients are coprime to <A>p</A>,
##  we use the fact that this ring homomorphism is defined via the Conway
##  polynomials on the <M>m</M>-th roots of unity, for <M>m</M> coprime to
##  <A>p</A>,
##  and is defined by mapping all <A>p</A><C>^</C><M>k</M>-th roots of unity
##  to the identity element of the finite field.
##  <P/>
##  This means that for <M>q =</M> <A>p</A><M>^k</M> and
##  <M>n = m q</M>, with <M>m</M> coprime to <A>p</A>,
##  the image of <M>\zeta_n^i</M> is equal to the image of <M>\zeta_m^j</M>,
##  where <M>i \equiv j q \bmod m</M> holds.
##  <P/>
##  <Example>
##  gap> FrobeniusCharacterValue( E(4), 2 );
##  fail
##  gap> FrobeniusCharacterValueExt( E(4), 2 );
##  Z(2)^0
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareGlobalFunction( "FrobeniusCharacterValueExt" );


#############################################################################
##
#A  CentralCharacter( <B> )
#O  CentralCharacter( <tbl>, <p>, <b> )
##
##  <#GAPDoc Label="CentralCharacter">
##  <ManSection>
##  <Attr Name="CentralCharacter" Arg="B"
##   Label="for a block object"/>
##  <Oper Name="CentralCharacter" Arg="tbl, p, b"
##   Label="for character table, characteristic, and block number"/>
##
##  <Description>
##  The <E>central character</E> of the <A>p</A>-block <M>B</M> of the group
##  <M>G</M> is the algebra homomorphism
##  <M>\lambda_B: &Centre;(FG) \rightarrow F</M> that is defined by 
##  <M>\lambda_B(C^+) = \omega_{{\chi}}(C^+)^*</M>,
##  for <M>\chi \in &Irr;(B)</M> and class sums <M>C^+</M> of <M>G</M>,
##  see <Cite Key="LP10" Where="p. 324"/>
##  or <Cite Key="Nav98" Where="p. 48"/>.
##  Here <M>\omega_{{\chi}}</M> is the central character of <M>\chi</M> as
##  defined in <Ref Oper="CentralCharacter" BookName="ref"/>, that is,
##  <M>\omega_{{\chi}}((g^G)^+) = |g^G| \chi(g) / \chi(1)</M> for
##  <M>g \in G</M>.
##  (Note that the values of <M>\omega_{{\chi}}</M> are complex numbers,
##  not elements of a finite field.)
##  <P/>
##  We represent the central character of the <A>b</A>-th <A>p</A>-block of
##  the character table <A>tbl</A> by the vector of reductions of the list
##  returned by <Ref Oper="CentralCharacter" BookName="ref"/> with argument
##  an irreducible character in the given block.
##  Note that the central character of the block depends on the choice of
##  the <A>p</A>-modular system, see Section <Ref Sect="sect:theor"/>.
##  <P/>
##  <Ref Oper="CentralCharacter" Label="for a block object"/> is based on
##  <Ref Func="FrobeniusCharacterValueExt"/>,
##  thus the returned list contains <K>fail</K> entries exactly when a
##  Conway polynomial would be needed that is not available.
##  <P/>
##  <Example>
##  gap> CentralCharacter( CharacterTable( "A5" ), 2, 1 );
##  [ Z(2)^0, Z(2)^0, 0*Z(2), 0*Z(2), 0*Z(2) ]
##  gap> CentralCharacter( Block( CharacterTable( "A5" ), 2, 1 ) );
##  [ Z(2)^0, Z(2)^0, 0*Z(2), 0*Z(2), 0*Z(2) ]
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareAttribute( "CentralCharacter", IsBlockOfCharacterTable );

DeclareOperation( "CentralCharacter",
    [ IsOrdinaryTable, IsPosInt, IsPosInt ] );


#############################################################################
##
#A  ClassPositionsOfKernel( <B> )
##
##  <#GAPDoc Label="ClassPositionsOfKernel">
##  <ManSection>
##  <Attr Name="ClassPositionsOfKernel" Arg="B" Label="for a block object"/>
##
##  <Description>
##  Let <A>B</A> be a <M>p</M>-block of the group <M>G</M>, say.
##  We define the <E>kernel</E> of <A>B</A> as the intersection of the
##  kernels of the ordinary irreducible characters in <A>B</A>.
##  and <Ref Attr="ClassPositionsOfKernel" Label="for a block object"/>
##  returns the list of positions of those conjugacy classes of <M>G</M>
##  (w. r. t. the underlying character table of <A>B</A>) that lie inside
##  the kernel.
##  <P/>
##  The kernel of <A>B</A> is the kernel of the natural action of <M>G</M>
##  on the algebra <M>e FG</M>,
##  where <M>e</M> is the block idempotent of <A>B</A>
##  (see <Ref Attr="CoefficientsOfOsimaIdempotent"
##  Label="for a block object"/>).
##  <P/>
##  The kernel of <A>B</A> is equal to the intersection of the kernel of
##  <E>any</E> character in <A>B</A> with the largest normal
##  <M>p'</M>-subgroup of <M>G</M>,
##  which is usually denoted by <M>O_{{p'}}(G)</M>,
##  see <Cite Key="LP10" Where="Theorem 4.11.13"/>,
##  <Cite Key="Nav98" Where="Theorem 6.10"/>, or
##  <Cite Key="NT89" Where="Ch. 5, Sect. 8"/>.
##  In particular, the kernel of the principal <M>p</M> block of <M>G</M>
##  is equal to <M>O_{{p'}}(G)</M>.
##  <P/>
##  <Example>
##  gap> t:= CharacterTable( "6.A6" );;
##  gap> List( PBlocks( t, 2 ), ClassPositionsOfKernel );
##  [ [ 1, 3, 5 ], [ 1, 3, 5 ], [ 1, 3, 5 ], [ 1 ], [ 1 ] ]
##  gap> List( PBlocks( t, 3 ), ClassPositionsOfKernel );
##  [ [ 1, 4 ], [ 1, 4 ], [ 1 ] ]
##  gap> List( PBlocks( t, 5 ), ClassPositionsOfKernel );
##  [ [ 1, 2, 3, 4, 5, 6 ], [ 1, 2, 3, 4, 5, 6 ], [ 1, 2, 3, 4, 5, 6 ], 
##    [ 1, 2, 3, 4, 5, 6 ], [ 1, 3, 5 ], [ 1, 3, 5 ], [ 1, 3, 5 ], 
##    [ 1, 4 ], [ 1, 4 ], [ 1, 4 ], [ 1, 4 ], [ 1 ], [ 1 ] ]
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareAttribute( "ClassPositionsOfKernel", IsBlockOfCharacterTable );


#############################################################################
##
#P  IsFaithful( <B> )
##
##  <#GAPDoc Label="IsFaithful">
##  <ManSection>
##  <Prop Name="IsFaithful" Arg="B" Label="for a block object"/>
##
##  <Description>
##  The block <A>B</A> is called <E>faithful</E> if its kernel is trivial,
##  see <Ref Attr="ClassPositionsOfKernel" Label="for a block object"/>.
##  <P/>
##  <Example>
##  gap> t:= CharacterTable( "6.A6" );;
##  gap> List( PBlocks( t, 2 ), IsFaithful );
##  [ false, false, false, true, true ]
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareProperty( "IsFaithful", IsBlockOfCharacterTable );


#############################################################################
##
#A  CoefficientsOfOsimaIdempotent( <B> )
#O  CoefficientsOfOsimaIdempotent( <tbl>, <p>, <b> )
##
##  <#GAPDoc Label="CoefficientsOfOsimaIdempotent">
##  <ManSection>
##  <Attr Name="CoefficientsOfOsimaIdempotent" Arg="B"
##   Label="for a block object"/>
##  <Oper Name="CoefficientsOfOsimaIdempotent" Arg="tbl, p, b"
##   Label="for character table, characteristic, and block number"/>
##
##  <Description>
##  Let <M>B</M> be a <A>p</A>-block of the group <M>G</M>.
##  The <E>primitive idempotent</E> corresponding to <M>\chi \in &Irr;(G)</M>
##  is
##  <Display Mode="M">
##  e_{{\chi}} = (\chi(1) / |G|) \sum_{g} \chi(g^{{-1}}) g \in &Centre;(&CC;G),
##  </Display>
##  where the summation runs over the elements of <M>G</M>.
##  <P/>
##  The <E>Osima idempotent</E> of <M>B</M> is defined as
##  <M>f_B = \sum_{{\chi}} e_{{\chi}}</M>,
##  where the summation runs over <M>\chi \in &Irr;(B)</M>.
##  The coefficient of <M>g \in G</M> in <M>f_B</M> is
##  <Display Mode="M">
##  a_g = (1 / |G|) \sum_{{\chi}} \chi(1) \chi(g^{{-1}}),
##  </Display>
##  where the summation runs over <M>\chi \in &Irr;(B)</M>.
##  <P/>
##  <Ref Attr="CoefficientsOfOsimaIdempotent" Label="for a block object"/>
##  returns the vector <M>[ a_{{g_1}}, a_{{g_2}}, \ldots ]</M>,
##  where <M>g_i</M> lies in the
##  <M>i</M>-th conjugacy class of the character table <A>tbl</A>.
##  <P/>
##  The name <Q>Osima idempotent</Q> appears in
##  <Cite Key="Isa76" Where="p. 277"/>.
##  Both <M>f_B</M> and its reduction <M>(f_B)^*</M> are called
##  <E>block idempotent</E><Index>block idempotent</Index>
##  in <Cite Key="LP10" Where="p. 324"/>,
##  whereas this term is used only for <M>e_B = (f_B)^*</M> 
##  in <Cite Key="Nav98" Where="p. 55"/>.
##  <P/>
##  <Example>
##  gap> CoefficientsOfOsimaIdempotent( CharacterTable( "A5" ), 2, 1 );
##  [ 11/15, 0, -1/15, 1/15, 1/15 ]
##  gap> CoefficientsOfOsimaIdempotent( Block( CharacterTable( "A5" ), 2, 1 ) );
##  [ 11/15, 0, -1/15, 1/15, 1/15 ]
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
DeclareAttribute( "CoefficientsOfOsimaIdempotent", IsBlockOfCharacterTable );

DeclareOperation( "CoefficientsOfOsimaIdempotent",
    [ IsOrdinaryTable, IsPosInt, IsPosInt ] );


#############################################################################
##
#E

