#############################################################################
##
#W  blockobj.gi          GAP 4 package CTBlocks                 Thomas Breuer
##
#Y  Copyright (C)  2013,   Lehrstuhl D für Mathematik,  RWTH Aachen,  Germany
##


#############################################################################
##
#V  BlocksFamily
##
BindGlobal( "BlocksFamily",
    NewFamily( "BlocksFamily", IsBlockOfCharacterTable ) );

BlocksFamily!.defaultType:= NewType( BlocksFamily,
    IsBlockOfCharacterTable and IsAttributeStoringRep );


#############################################################################
##
#M  \=( <B1>, <B2> )
#M  \<( <B1>, <B2> )
##
##  <#GAPDoc Label="compare_blocks">
##  Two objects that represent blocks of character tables are compared w.r.t.
##  the triples <C>[ id, p, b ]</C>,
##  where <C>id</C> is the identifier of the underlying character table,
##  <C>p</C> is the underlying characteristic,
##  and <C>b</C> is the number of the block.
##  <#/GAPDoc>
##
InstallMethod( \=,
    [ "IsBlockOfCharacterTable", "IsBlockOfCharacterTable" ],
    function( B1, B2 )
    return [ Identifier( UnderlyingCharacterTable( B1 ) ),
             UnderlyingCharacteristic( B1 ),
             NumberOfBlock( B1 ) ] =
           [ Identifier( UnderlyingCharacterTable( B2 ) ),
             UnderlyingCharacteristic( B2 ),
             NumberOfBlock( B2 ) ];
    end );

InstallMethod( \<,
    [ "IsBlockOfCharacterTable", "IsBlockOfCharacterTable" ],
    function( B1, B2 )
    return [ Identifier( UnderlyingCharacterTable( B1 ) ),
             UnderlyingCharacteristic( B1 ),
             NumberOfBlock( B1 ) ] <
           [ Identifier( UnderlyingCharacterTable( B2 ) ),
             UnderlyingCharacteristic( B2 ),
             NumberOfBlock( B2 ) ];
    end );


#############################################################################
##
#M  String( <B> )
#M  ViewString( <B> )
#M  DisplayString( <B> )
##
##  Provide everything for printing blocks,
##  see ?Recommendations for the implementation.
##
Perform( [ String, ViewString, PrintString, DisplayString ],

         function( op ) CallFuncList( InstallMethod, [ op, 
    [ "IsBlockOfCharacterTable" ],
    B -> Concatenation( "Block( CharacterTable( \"",
             Identifier( UnderlyingCharacterTable( B ) ), "\" ), ",
             String( UnderlyingCharacteristic( B ) ), ", ",
             String( NumberOfBlock( B ) ), " )" ) ] ); end );


#############################################################################
##
#M  Block( <tbl>, <p>, <b> )
##
InstallMethod( Block,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    local bl, obj;

    bl:= ComputedPBlocks( tbl );

    # Perhaps the block object is already stored.
    if IsBound( bl.( p ) ) then
      bl:= bl.( p );
      if IsBound( bl[b] ) then
        return bl[b];
      fi;
    else
      bl.( p ):= [];
      bl:= bl.( p );
    fi;

    # The block object must be created.
    if Length( PrimeBlocks( tbl, p ).defect ) < b then
      return fail;
    fi;

    obj:= rec();
    ObjectifyWithAttributes( obj,
        BlocksFamily!.defaultType,
        UnderlyingCharacterTable, tbl,
        UnderlyingCharacteristic, p,
        NumberOfBlock, b );
#T Why does ObjectifyWithAttributes not return the object?

    bl[b]:= obj;

    return obj;
    end );


#############################################################################
##
#M  ComputedPBlocks( <tbl> )
##
InstallMethod( ComputedPBlocks,
    [ "IsOrdinaryTable" ],
    tbl -> rec() );


#############################################################################
##
#F  PRegularTable( <tbl>, <p> )
##
InstallGlobalFunction( PRegularTable, function( tbl, p )
    local cache;

    cache:= ComputedPRegularTables( tbl );
    if not IsBound( cache.( p ) ) then
      cache.( p ):= CharacterTableRegular( tbl, p );
    fi;

    return cache.( p );
    end );


#############################################################################
##
#M  ComputedPRegularTables( <tbl> )
##
InstallMethod( ComputedPRegularTables,
    [ "IsOrdinaryTable" ],
    tbl -> rec() );


#############################################################################
##
#F  PBlocks( <tbl>, <p> )
##
InstallGlobalFunction( PBlocks,
    function( arg )
    if Length( arg ) = 2 and
       IsOrdinaryTable( arg[1] ) and
       IsInt( arg[2] ) and IsPrimeInt( arg[2] ) then

      # Compute the <p>-blocks that are not yet stored.
      CallFuncList( PBlocksOp, arg );
      return ComputedPBlocks( arg[1] ).( arg[2] );
    else
      Error( "usage: ",
        "PBlocks(<tbl>,<p>)",
        " for ordinary character table <tbl>, and prime integer <p>" );
    fi;
    end );


#############################################################################
##
#M  PBlocksOp( <tbl>, <p> )
##  
InstallMethod( PBlocksOp,
    [ "IsOrdinaryTable", "IsPosInt" ],
    function( tbl, p )
    return List( [ 1 .. Length( PrimeBlocks( tbl, p ).defect ) ],
                 b -> Block( tbl, p, b ) );
    end );


#############################################################################
##
#M  PrincipalBlock( <tbl>, <p> )
##
##  Note that the trivial character need not be the first in the list of
##  irreducibles of <tbl>.
##
InstallMethod( PrincipalBlock,
    [ "IsOrdinaryTable", "IsPosInt" ],
    function( tbl, p )
    local bl, irr, i;

    bl:= PrimeBlocks( tbl, p );
    irr:= Irr( tbl );
    for i in [ 1 .. Length( irr ) ] do
      if irr[i][1] = 1 and
         Set( ValuesOfClassFunction( irr[i] ) ) = [ 1 ] then
        return Block( tbl, p, bl.block[i] );
      fi;
    od;
    Error( "<tbl> has no trivial character?" );
    end );


#############################################################################
##
#M  DecompositionMatrix( <B> )
##
InstallMethod( DecompositionMatrix,
    [ "IsBlockOfCharacterTable" ],
    function( B )
    local modtbl;

    if Defect( B ) = 0 then
      return [ [ 1 ] ];
    fi;

    modtbl:= UnderlyingCharacterTable( B ) mod UnderlyingCharacteristic( B );
    if modtbl = fail then
      return fail;
    fi;
    return DecompositionMatrix( modtbl, NumberOfBlock( B ) );
    end );


#############################################################################
##
#M  DimensionsOfDecompositionMatrix( <B> )
##
InstallMethod( DimensionsOfDecompositionMatrix,
    [ "IsBlockOfCharacterTable" ],
    function( B )
    local tbl, p, orders, pregular, i, bl, irr;

    if Defect( B ) = 0 then
      return [ 1, 1 ];
    fi;

    tbl:= UnderlyingCharacterTable( B );
    p:= UnderlyingCharacteristic( B );
    orders:= OrdersClassRepresentatives( tbl );
    pregular:= [ 1 ];
    for i in [ 2 .. Length( orders ) ] do
      if orders[i] mod p <> 0 then
        Add( pregular, i );
      fi;
    od;

    bl:= PrimeBlocks( tbl, p );
    irr:= List( Irr( tbl ){ Positions( bl.block, NumberOfBlock( B ) ) },
                x -> ValuesOfClassFunction( x ){ pregular } );

    return [ Length( irr ), RankMat( IntegralizedMat( irr ).mat ) ];
    end );


#############################################################################
##
#M  Dimension( <B> )
##
InstallMethod( Dimension,
    [ "IsBlockOfCharacterTable" ],
    function( B )
    local result, chi;

    result:= 0;
    for chi in Irr( B ) do
      result:= result + chi[1]^2;
    od;
    return result;
    end );


#############################################################################
##
#F  FrobeniusCharacterValueExt( <value>, <p> )
##
InstallGlobalFunction( FrobeniusCharacterValueExt, function( value, p )
    local n, m, q, coeffs, newcoeffs, i, j;

    n:= Conductor( value );
    if n mod p = 0 then
      # We have to work.
      m:= n;
      q:= 1;
      while m mod p = 0 do
        m:= m / p;
        q:= q * p;
      od;

      coeffs:= COEFFS_CYC( value );
      newcoeffs:= ListWithIdenticalEntries( m, 0 );
      for i in [ 1 .. n ] do
        j:= ( ( i-1) / q ) mod m;
        newcoeffs[j+1]:= newcoeffs[j+1] + coeffs[i];
      od;
      value:= CycList( newcoeffs );
    fi;

    # Now the conductor is coprime to <p>.
    return FrobeniusCharacterValue( value, p );
    end );


#############################################################################
##
#M  CentralCharacter( <tbl>, <p>, <b> )
##
InstallMethod( CentralCharacter,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    local bl, pos;

    bl:= PrimeBlocks( tbl, p );
    pos:= Position( bl.block, b );
    if pos = fail then
      Error( "<tbl> does not have <b> blocks" );
    fi;
    return List( CentralCharacter( Irr( tbl )[ pos ] ),
                 x -> FrobeniusCharacterValueExt( x, p ) );
    end );


#############################################################################
##
#M  CentralCharacter( <B> )
##
InstallMethod( CentralCharacter,
    [ "IsBlockOfCharacterTable" ],
    B -> CentralCharacter( UnderlyingCharacterTable( B ),
             UnderlyingCharacteristic( B ), NumberOfBlock( B ) ) );


#############################################################################
##
#M  ClassPositionsOfKernel( <B> )
##
InstallMethod( ClassPositionsOfKernel, 
    [ "IsBlockOfCharacterTable" ],
    function( B )
    local tbl, result, chi;

    tbl:= UnderlyingCharacterTable( B );
    result:= [ 1 .. NrConjugacyClasses( tbl ) ];
    for chi in Irr( B ) do
      IntersectSet( result, ClassPositionsOfKernel( chi ) );
    od;

    return result;
    end );


#############################################################################
##
#M  IsFaithful( <B> )
##
InstallMethod( IsFaithful, 
    [ "IsBlockOfCharacterTable" ],
    B -> Length( ClassPositionsOfKernel( B ) ) = 1 );


#############################################################################
##
#F  CoefficientsOfOsimaIdempotent( <tbl>, <p>, <b> )
##
InstallMethod( CoefficientsOfOsimaIdempotent,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    local blocks, coeffs, irr, found, i, chi;

    blocks:= PrimeBlocks( tbl, p );
    coeffs:= 0 * [ 1 .. NrConjugacyClasses( tbl ) ];
    irr:= List( Irr( tbl ), ValuesOfClassFunction );
    found:= false;
    for i in [ 1 .. Length( blocks.block ) ] do
      if blocks.block[i] = b then
        found:= true;
        chi:= irr[i];
        coeffs:= coeffs + chi[1] * chi;
      fi;
    od;

    if found then
      return ComplexConjugate( coeffs ) / Size( tbl );
    else
      return fail;
    fi;
    end );


#############################################################################
##
#F  CoefficientsOfOsimaIdempotent( <B> )
##
InstallMethod( CoefficientsOfOsimaIdempotent,
    [ "IsBlockOfCharacterTable" ],
    B -> CoefficientsOfOsimaIdempotent( UnderlyingCharacterTable( B ),
             UnderlyingCharacteristic( B ), NumberOfBlock( B ) ) );


#############################################################################
##
#M  Defect( <tbl>, <p>, <b> )
##
InstallMethod( Defect,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    local bl;

    bl:= PrimeBlocks( tbl, p );
    if b <= Length( bl.defect ) then
      return bl.defect[b];
    fi;
    Error( "<tbl> does not have <b> blocks" );
    end );


#############################################################################
##
#M  Defect( <B> )
##
InstallMethod( Defect,
    [ "IsBlockOfCharacterTable" ],
    B -> PrimeBlocks( UnderlyingCharacterTable( B ),
             UnderlyingCharacteristic( B ) ).defect[ NumberOfBlock( B ) ] );


#############################################################################
##
#M  Irr( <ordtbl>, <p>, <b> )
##
InstallMethod( Irr,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( ordtbl, p, b )
    return Irr( ordtbl ){ Positions( PrimeBlocks( ordtbl, p ).block, b ) };
    end );


#############################################################################
##
#M  Irr( <B> )
##
InstallMethod( Irr,
    [ "IsBlockOfCharacterTable" ],
    function( B )
    local ordtbl, p, bl;

    ordtbl:= UnderlyingCharacterTable( B );
    p:= UnderlyingCharacteristic( B );
    bl:= PrimeBlocks( ordtbl, p );

    return Irr( ordtbl ){ Positions( bl.block, NumberOfBlock( B ) ) };
    end );


#############################################################################
##
#M  IBr( <modtbl>, <b> )
##
InstallMethod( IBr,
    [ "IsBrauerTable", "IsPosInt" ],
    function( modtbl, b )
    local bl;

    bl:= BlocksInfo( modtbl );
    if b <= Length( bl ) then
      return Irr( modtbl ){ bl[b].modchars };
    fi;
    Error( "<modtbl> does not have <b> blocks" );
    end );


#############################################################################
##
#M  IBr( <B> )
##
##  We can compute the irreducibles in the following situations:
##
##  - The Brauer table in question (with irreducibles) is known.
##  - The block has defect zero.
##  - The block has cyclic defect and all its irreducible Brauer characters
##    lift to ordinary characters.
##
InstallMethod( IBr,
    [ "IsBlockOfCharacterTable" ],
    function( B )
    local ordtbl, p, b, modtbl, bl, chi, inv, brest, brestset, l;

    ordtbl:= UnderlyingCharacterTable( B );
    p:= UnderlyingCharacteristic( B );
    b:= NumberOfBlock( B );
    modtbl:= ordtbl mod p;
    if modtbl <> fail then
      bl:= BlocksInfo( modtbl );
      return Irr( modtbl ){ bl[b].modchars };
    elif Defect( B ) = 0 then
      bl:= PrimeBlocks( ordtbl, p );
      modtbl:= PRegularTable( ordtbl, p );
      chi:= Irr( ordtbl )[ Position( bl.block, b ) ];
      return [ RestrictedClassFunction( chi, modtbl ) ];
    elif Defect( B ) = 1 then
      # If we are lucky, all Brauer characters lift to characteristic zero.
      modtbl:= PRegularTable( ordtbl, p );
      bl:= PrimeBlocks( ordtbl, p );
      inv:= InverseMap( bl.block );
      brest:= RestrictedClassFunctions( Irr( ordtbl ){ inv[b] }, modtbl );
      brestset:= Set( brest );
      l:= Length( brestset );
      if l = 1 then
        # There is only one irreducible Brauer character in the block.
        return brestset;
      else
        if Sum( brestset{ [ 1 .. l-1 ] } ) = brestset[l] then
          # All Brauer characters lift to ordinary characters.
          # Sort the Brauer characters like the ordinary lifts.
          return brest{ SortedList( List( [ 1 .. l-1 ],
                            i -> Position( brest, brestset[i] ) ) ) };
        fi;
      fi;
    fi;

    # We have no criterion.
    return fail;
    end );


#############################################################################
##
#E

