#############################################################################
##
#W  blocks.gi            GAP 4 package CTBlocks                 Thomas Breuer
##
#Y  Copyright (C)  2013,   Lehrstuhl D für Mathematik,  RWTH Aachen,  Germany
##


#############################################################################
##
#M  IsRadicalPSubgroup( <G>, <D>, <p> )
##
InstallMethod( IsRadicalPSubgroup,
    [ "IsGroup", "IsGroup", "IsPosInt" ],
    function( G, D, p )
    local Dsize, coll, N, index;

    Dsize:= Size( D );
    if Dsize <> 1 then
      coll:= Collected( Factors( Dsize ) );
      if Length( coll ) <> 1 or coll[1][1] <> p then
        # <D> is not a <p>-group.
        return false;
      fi;
    fi;

    if ( Size( G ) / Dsize ) mod p <> 0 then
      # <D> is a Sylow <p>-subgroup.
      return true;
    fi;

    N:= Normalizer( G, D );
    index:= Size( N ) / Dsize;
    if not IsInt( index ) then
      # <D> cannot be a subgroup of <G>, give a useful error message.
      Error( "<D> is not a subgroup of <G>" );
    elif index mod p <> 0 then
      # N_{<G>}(<D>) / <D> is a <p>'-group.
      SetPCore( N, p, D );
      return true;
    fi;

    return Size( PCore( N, p ) ) = Dsize;
    end );


#############################################################################
##
#M  NormalizerOfRadicalPSubgroup( <G>, <D>, <p> )
##
InstallMethod( NormalizerOfRadicalPSubgroup,
    [ "IsGroup", "IsGroup", "IsPosInt" ],
    function( G, D, p )
    local Dsize, coll, N, index;

    Dsize:= Size( D );
    if Dsize <> 1 then
      coll:= Collected( Factors( Dsize ) );
      if Length( coll ) <> 1 or coll[1][1] <> p then
        # <D> is not a <p>-group.
        return fail;
      fi;
    fi;

    N:= Normalizer( G, D );
    if ( Size( G ) / Dsize ) mod p <> 0 then
      # <D> is a Sylow <p>-subgroup in <G> and thus radical.
      # (Computing the order of <G> is expected to be cheaper than
      # computing the order of 'N'.)
      SetPCore( N, p, D );
      return N;
    fi;

    index:= Size( N ) / Dsize;
    if not IsInt( index ) then
      # <D> cannot be a subgroup of <G>, give a useful error message.
      Error( "<D> is not a subgroup of <G>" );
    elif index mod p <> 0 then
      # N_{<G>}(<D>) / <D> is a <p>'-group.
      SetPCore( N, p, D );
      return N;
    elif Size( PCore( N, p ) ) = Dsize then
      # <D> is <p>-radical.
      return N;
    fi;

    return fail;
    end );


#############################################################################
##
#M  ComputedRepresentativesRadicalPSubgroupsAndNormalizers( <G> )
#M  ComputedRepresentativesRadicalPSubgroupsAndNormalizers( <tom> )
##
InstallMethod( ComputedRepresentativesRadicalPSubgroupsAndNormalizers,
    [ "IsGroup" ],
    G -> rec() );

InstallMethod( ComputedRepresentativesRadicalPSubgroupsAndNormalizers,
    [ "IsTableOfMarks and IsTableOfMarksWithGens" ],
    tom -> rec() );


#############################################################################
##
#F  RepresentativesRadicalPSubgroupsAndNormalizers( <G>, <p>[, <arec>] )
#F  RepresentativesRadicalPSubgroupsAndNormalizers( <tom>, <p>[, <arec>] )
##
InstallGlobalFunction( RepresentativesRadicalPSubgroupsAndNormalizers,
    function( arg )
    local p, stored;

    if Length( arg ) in [ 2, 3 ] and
       ( IsGroup( arg[1] ) or IsTableOfMarksWithGens( arg[1] ) ) and
       IsInt( arg[2] ) and IsPrimeInt( arg[2] ) and
       ( Length( arg ) = 2 or IsRecord( arg[3] ) ) then
      p:= arg[2];
      stored:= ComputedRepresentativesRadicalPSubgroupsAndNormalizers(
                   arg[1] );
      if not IsBound( stored.( p ) ) then
        stored.( p ):= CallFuncList(
                           RepresentativesRadicalPSubgroupsAndNormalizersOp,
                           arg );
      fi;
      return stored.( p );
    else
      Error( "usage: ",
        "RepresentativesRadicalPSubgroupsAndNormalizers(<G>,<p>[,<arec>])",
        " for group or table of marks <G>, and prime integer <p>" );
    fi;
    end );


#############################################################################
##
#M  RepresentativesRadicalPSubgroupsAndNormalizersOp( <G>, <p>[, <arec>] )
##
InstallMethod( RepresentativesRadicalPSubgroupsAndNormalizersOp,
    [ "IsGroup", "IsPosInt" ],
    function( G, p )
    local reps, core, syl, epi, img, iso, ccl, n, i, U, N;

    Info( InfoBlocks, 1,
          "RepresentativesRadicalPSubgroupsAndNormalizersOp:\n",
          "#I  use the simpleminded approach for group\n",
          "#I  ", ViewString( G ) );

    reps:= [];

    # - Work in the factor group modulo the <p>-core.
    #   (Note that this is compatible with normalizers and conjugacy.)
    # - Inspect conjugacy classes of subgroups of a Sylow <p>-subgroup.
    #   (This is a very simpleminded approach.)
    core:= PCore( G, p );
    if 1 < Size( core ) then
      Info( InfoBlocks, 2,
            "factor out ", p, "-core of order ", Size( core ) );
    fi;
    epi:= NaturalHomomorphismByNormalSubgroup( G, core );
    img:= Images( epi );
    syl:= SylowSubgroup( img, p );
    iso:= IsomorphismPcGroup( syl );
    ccl:= ConjugacyClassesSubgroups( Images( iso ) );
    n:= Length( ccl );
    for i in [ 1 .. n ] do
      U:= PreImages( iso, Representative( ccl[i] ) );
      N:= NormalizerOfRadicalPSubgroup( img, U, p );
      if N <> fail and
         ForAll( reps, r -> not IsConjugate( img, r.subgroup, U ) ) then
        Add( reps, rec( subgroup:= U, normalizer:= N ) );
      fi;
      Info( InfoBlocks, 2,
            "RepresentativesRadicalPSubgroupsAndNormalizersOp: ",
            "done ", i, " of ", n, ", have ", Length( reps ) );
    od;

    return List( reps,
                 r -> rec( subgroup:= PreImages( epi, r.subgroup ),
                           normalizer:= PreImages( epi, r.normalizer ) ) );
    end );

InstallMethod( RepresentativesRadicalPSubgroupsAndNormalizersOp,
    [ "IsGroup", "IsPosInt", "IsRecord" ],
    function( G, p, arec )
    local reps, mreps, m, rho, img, core, epi, pair, U, N,
          syl, ccl, n, i;

    if not IsBound( arec.Maxes ) then
      # We use the simpleminded approach.
      return RepresentativesRadicalPSubgroupsAndNormalizers( G, p );
    fi;

    Info( InfoBlocks, 1,
          "RepresentativesRadicalPSubgroupsAndNormalizersOp: ",
          "compute in maximal subgroups" );

    # Start with the p-core of the big group.
    if IsBound( arec.CharacterTable ) and
       ClassPositionsOfPCore( arec.CharacterTable, p ) = [ 1 ] then
      reps:= [ rec( subgroup:= TrivialSubgroup( G ), normalizer:= G ) ];
    else
      reps:= [ rec( subgroup:= PCore( G, p ), normalizer:= G ) ];
    fi;

    # Assume that `arec.Maxes' covers the relevant classes of subgroups.
    for m in arec.Maxes do
      if Size( m ) mod p = 0 then

        Info( InfoBlocks, 2,
              "RepresentativesRadicalPSubgroupsAndNormalizersOp: ",
              "call for given subgroup ", ViewString( m ) );

        # Switch to a better representation if possible.
        if IsPermGroup( m ) then
          rho:= SmallerDegreePermutationRepresentation( m );
        else
          rho:= IdentityMapping( m );
        fi;
        img:= Images( rho, m );

        # Factor out the p-core.
        core:= PCore( img, p );
        epi:= NaturalHomomorphismByNormalSubgroup( img, core );
        img:= Images( epi );

        # Run over the radical p-subgroups of the subquotient.
        for pair in RepresentativesRadicalPSubgroupsAndNormalizers(
                        img, p ) do
          U:= PreImages( epi, pair.subgroup );
          U:= PreImages( rho, U );
# if the trivial subgroup was lifted to a nontrivial group then
# add [ U, m ] to the safe list: U is p-radical (because the normalizer in G
# cannot be larger) and cannot be conjugate to a pair obtained from another
# max. subgroup!
# carry over also the normalizers? (as a lower bound for the norm. in G)
          N:= NormalizerOfRadicalPSubgroup( G, U, p );
          if N <> fail then
#T omit pairs with the trivial subgroup!
            # Check whether we have found a new representative.
            # (If necessary then check conjugacy in the big group,
            # but check the order of subgroup and normalizer first.)
            if ForAll( reps,
                       x -> Size( x.subgroup ) <> Size( U ) or
                            Size( x.normalizer ) <> Size( N ) or
                            not IsConjugate( G, x.subgroup, U ) ) then
              Add( reps, rec( subgroup:= U, normalizer:= N ) );
            fi;
          fi;
        od;
      fi;
    od;

    return reps;
    end );


#############################################################################
##
#M  RepresentativesRadicalPSubgroupsAndNormalizersOp( <tom>, <p>[, <arec>] )
##
##  In the case of a table of marks <tom>, we ignore <arec>.
##
InstallMethod( RepresentativesRadicalPSubgroupsAndNormalizersOp,
    [ "IsTableOfMarks and IsTableOfMarksWithGens", "IsPosInt" ],
    function( tom, p )
    local G, core, div, orders, n, reps, i, U, N;

    # We know class representatives and their orders.
    G:= UnderlyingGroup( tom );
    core:= Size( PCore( G, p ) );
    if core = 1 then
      div:= p;
    else
      div:= core;
    fi;
    orders:= OrdersTom( tom );
    n:= Length( orders );
    reps:= [];
    for i in [ 1 .. n ] do
      if orders[i] = 1 then
        if core = 1 then
          Add( reps, rec( subgroup:= TrivialSubgroup( G ),
                          normalizer:= G,
                          subgroupnr:= i,
                          normalizernr:= Position( orders, Size( G ) ) ) );
        fi;
      elif orders[i] mod div = 0 and IsPrimePowerInt( orders[i] ) then
        U:= RepresentativeTom( tom, i );
        N:= NormalizerOfRadicalPSubgroup( G, U, p );
        if N <> fail then
          Add( reps, rec( subgroup:= U,
                          normalizer:= N,
                          subgroupnr:= i,
                          normalizernr:= NormalizersTom( tom )[i] ) );
        fi;
      fi;
      Info( InfoBlocks, 2,
            "RepresentativesRadicalPSubgroupsAndNormalizersOp: ",
            "done ", i, " of ", n, ", have ", Length( reps ) );
    od;

    return reps;
    end );

InstallMethod( RepresentativesRadicalPSubgroupsAndNormalizersOp,
    [ "IsTableOfMarks and IsTableOfMarksWithGens", "IsPosInt", "IsRecord" ],
    function( tom, p, arec )
    return RepresentativesRadicalPSubgroupsAndNormalizers( tom, p );
    end );


#############################################################################
##
#M  ComputedRepresentativesChainsOfRadicalPSubgroupsAndNormalizers( <G> )
#M  ComputedRepresentativesChainsOfRadicalPSubgroupsAndNormalizers( <tom> )
##
InstallMethod( ComputedRepresentativesChainsOfRadicalPSubgroupsAndNormalizers,
    [ "IsGroup" ],
    G -> rec() );

InstallMethod( ComputedRepresentativesChainsOfRadicalPSubgroupsAndNormalizers,
    [ "IsTableOfMarks and IsTableOfMarksWithGens" ],
    tom -> rec() );


#############################################################################
##
#F  RepresentativesChainsOfRadicalPSubgroupsAndNormalizers(
#F      <G>, <p>[, <arec>] )
#F  RepresentativesChainsOfRadicalPSubgroupsAndNormalizers(
##      <tom>, <p>[, <arec>] )
##
##  Take the value if it is stored, and delegate to the operation
##  'RepresentativesChainsOfRadicalPSubgroupsAndNormalizersOp' otherwise.
##
InstallGlobalFunction( RepresentativesChainsOfRadicalPSubgroupsAndNormalizers,
    function( arg )
    local p, stored;

    if Length( arg ) in [ 2, 3 ] and
       ( IsGroup( arg[1] ) or IsTableOfMarksWithGens( arg[1] ) ) and
       IsInt( arg[2] ) and IsPrimeInt( arg[2] ) and
       ( Length( arg ) = 2 or IsRecord( arg[3] ) ) then
      p:= arg[2];
      stored:= ComputedRepresentativesChainsOfRadicalPSubgroupsAndNormalizers(
                   arg[1] );
      if not IsBound( stored.( p ) ) then
        stored.( p ):= CallFuncList(
            RepresentativesChainsOfRadicalPSubgroupsAndNormalizersOp,
                           arg );
      fi;
      return stored.( p );
    else
      Error( "usage: ",
        "RepresentativesChainsOfRadicalPSubgroupsAndNormalizers(",
        "<G>,<p>[,<arec>])",
        " for group or table of marks <G>, and prime integer <p>" );
    fi;
    end );


#############################################################################
##
#M  RepresentativesChainsOfRadicalPSubgroupsAndNormalizersOp(
#M      <G>, <p>[, <arec>] )
##
##  In the case that <arec> is not given, compute the radical <p>-subgroups
##  and delegate to the case with <arec>;
##  this is the method that really does the work:
##
##  - The <p>-chains in question of length one are given by the chains
##    $O_{<p>}(<G>) < P$, where $P$ runs over a set of representatives of
##    radical <p>-subgroups of <G>.
##
##  - For a <p>-chain $C: O_{<p>}(<G>) < P_1 < \cdots < P_k$ and $P_{k+1}$
##    with the property $P_k < P_{k+1}$, we define the extension of $C$ with
##    $P_{k+1}$ as the chain
##    $C < P_{k+1}: O_{<p>}(<G>) < P_1 < \cdots < P_k < P_{k+1}$.
##    For $g \in G$, the <p>-chain $C^g$ is defined as
##    $O_{<p>}(<G>) < P_1^g < \cdots < P_k^g$.
##
##  - If class representatives of the relevant <p>-chains of length $k$ are
##    known then class representatives of the relevant <p>-chains of length
##    $k+1$ are given by the extensions $C^g < P$ where $C$ runs over the
##    known relevant <p>-chains of length $k$, $P$ runs over the given
##    representatives of radical <p>-subgroups of <G>, and $g$ runs over
##    a set of representatives of those $N_G(C)$-$N_G(P)$-double cosets
##    for which $C^g < P$ is a <p>-chain.
##
##    Proof:
##    We have to show that (a) any two of the above extensions are not
##    <G>-conjugate and (b) that any extension is <G>-conjugate to one of the
##    above extensions.
##
##    Ad (a):
##    Let $C_1 < P_1$ and $C_2 < P_2$ be two <p>-chains of length $k+1$ that
##    occur in the above set and that are <G>-conjugate, that is,
##    $C_1^g = C_2$ and $P_1^g = P_2$ hold for some $g \in G$.
##    Then $P_1 = P_2$ holds, that is, $g \in N_G(P_1)$.
##    Moreover, $C_1 = C^{g_1}$ and $C_2 = C^{g_2}$ for some representative
##    $C$ of relevant chains of length $k$ and $g_1, g_2 \in G$.
##    Thus $g_1 g g_2^{-1} \in N_G(C)$, that is,
##    $g_1 \in N_G(C) g_2 N_G(P_1)$, which implies that $g_1 = g_2$.
##
##    Ad (b):
##    Let $C < P$ be a relevant <p>-chain of length $k+1$.
##    Without loss of generality, let $P$ be one of the given representatives
##    of radical <p>-subgroups of <G>, and let $C = C_0^h$ for some
##    representative $C_0$ of relevant <p>-chains of length $k$,
##    and $h \in G$.
##    Let $g$ be the representative of $N_G(C_0)$-$N_G(P)$-double cosets
##    for which $h \in N_G(C_0) g N_G(P)$ holds.
##    Then $C < P$ is <G>-conjugate to $C_0^g < P$.
##
InstallMethod( RepresentativesChainsOfRadicalPSubgroupsAndNormalizersOp,
    [ "IsGroup", "IsPosInt" ],
    function( G, p )
    return RepresentativesChainsOfRadicalPSubgroupsAndNormalizersOp(
               G, p, rec( radicalsubgroups:=
               RepresentativesRadicalPSubgroupsAndNormalizers( G, p ) ) );
    end );

InstallMethod( RepresentativesChainsOfRadicalPSubgroupsAndNormalizersOp,
    [ "IsGroup", "IsPosInt", "IsRecord" ],
    function( G, p, arec )
    local grps, chains, sizesbyp, new, k, ch, P, N, s, gens, pos, cand, pair;

    if not IsBound( arec.radicalsubgroups ) then
      arec.radicalsubgroups:=
        RepresentativesRadicalPSubgroupsAndNormalizers( G, p, arec );
    fi;

    # Compute class representatives of radical p-subgroups.
    grps:= arec.radicalsubgroups;
    if Length( grps ) = 1 then
      return [];
    fi;

    # We omit the p-core, this yields the chains of length one.
    chains:= [];
    grps:= grps{ [ 2 .. Length( grps ) ] };
    sizesbyp:= List( grps, r -> Size( r.subgroup ) / p );
    new:= List( grps, r -> rec( subgroups:= [ r.subgroup ],
                                normalizer:= r.normalizer ) );
    k:= 0;

    repeat

      # Given the chains of length k, compute the chains of length k+1.
      k:= k+1;
      chains[k]:= new;
      new:= [];

      # Run over the chains of length k.
      for ch in chains[k] do

        P:= ch.subgroups[k];
        N:= ch.normalizer;
        s:= Size( P );
        gens:= GeneratorsOfGroup( P );

        # Run over the representatives that may extend the chain.
        for pos in PositionsProperty( sizesbyp, x -> x mod s = 0 ) do
          cand:= grps[ pos ];

          # Run over double coset representatives of the normalizers.
          for pair in DoubleCosetRepsAndSizes( G, N, cand.normalizer ) do
            if ForAll( gens, x -> x^pair[1] in cand.subgroup ) then
              Add( new, rec(
                subgroups:= Concatenation(
                    List( ch.subgroups,
                          x -> ConjugateSubgroup( x, pair[1] ) ),
                    [ cand.subgroup ] ),
                normalizer:= Intersection(
                    ConjugateSubgroup( ch.normalizer, pair[1] ),
                    cand.normalizer ) ) );
            fi;
          od;
        od;
      od;

    until Length( new ) = 0;

    return chains;
    end );


#############################################################################
##
#M  RepresentativesChainsOfRadicalPSubgroupsAndNormalizersOp(
#M      <tom>, <p>[, <arec>] )
##
##  In the case of a table of marks <tom>, we use the table of marks
##  (and the argument <arec> if applicable) for computing representatives of
##  radical <p>-subgroups, and then delegate to the method for the group.
##
InstallMethod( RepresentativesChainsOfRadicalPSubgroupsAndNormalizersOp,
    [ "IsTableOfMarks and IsTableOfMarksWithGens", "IsPosInt" ],
    function( tom, p )
    return RepresentativesChainsOfRadicalPSubgroupsAndNormalizersOp(
               UnderlyingGroup( tom ), p, rec( radicalsubgroups:=
               RepresentativesRadicalPSubgroupsAndNormalizers( tom, p ) ) );
    end );

InstallMethod( RepresentativesChainsOfRadicalPSubgroupsAndNormalizersOp,
    [ "IsTableOfMarks and IsTableOfMarksWithGens", "IsPosInt", "IsRecord" ],
    function( tom, p, arec )
    arec:= ShallowCopy( arec );
    arec.radicalsubgroups:=
      RepresentativesRadicalPSubgroupsAndNormalizers( tom, p, arec );

    return RepresentativesChainsOfRadicalPSubgroupsAndNormalizersOp(
               UnderlyingGroup( tom ), p, arec );
    end );


#############################################################################
##
#M  ComputedDescriptionOfPWeights( <tbl> )
##
InstallMethod( ComputedDescriptionOfPWeights,
    [ "IsOrdinaryTable" ],
    tbl -> rec() );


#############################################################################
##
#F  DescriptionOfPWeights( <B> )
#F  DescriptionOfPWeights( <tbl>, <p>[, <b>] )
##
InstallGlobalFunction( DescriptionOfPWeights,
    function( arg )
    local p, stored;

    if Length( arg ) = 1 and IsBlockOfCharacterTable( arg[1] ) then
      return DescriptionOfPWeightsAttr( arg[1] );
    elif Length( arg ) = 2 and
       IsOrdinaryTable( arg[1] ) and
       IsInt( arg[2] ) and IsPrimeInt( arg[2] ) then
      p:= arg[2];
      stored:= ComputedDescriptionOfPWeights( arg[1] );
      if not IsBound( stored.( p ) ) then
        stored.( p ):= DescriptionOfPWeightsOp( arg[1], p );
      fi;
      return stored.( p );
    elif Length( arg ) = 3 and
       IsOrdinaryTable( arg[1] ) and
       IsInt( arg[2] ) and IsPrimeInt( arg[2] ) and IsPosInt( arg[3] ) then
      return CallFuncList( DescriptionOfPWeightsOp, arg );
    else
      Error( "usage: ",
        "DescriptionOfPWeights(<B>) for p-block <B> or ",
        "DescriptionOfPWeights(<tbl>,<p>[,<b>])",
        " for char. table <tbl>, prime integer <p>, and block number <b>" );
    fi;
    end );


#############################################################################
##
#M  DescriptionOfPWeightsAttr( <B> )
##
InstallMethod( DescriptionOfPWeightsAttr,
    [ "IsBlockOfCharacterTable" ],
    B -> DescriptionOfPWeightsOp( UnderlyingCharacterTable( B ),
             UnderlyingCharacteristic( B ),
             NumberOfBlock( B ) ) );


#############################################################################
##
#F  AddBlockDistributionToDescriptionOfPWeights( <tbl>, <p>, <r> )
##
##  This function can be used for computing the 'blocks' component,
##  independent of the method how the normalizers and characters are
##  computed.
##
AddBlockDistributionToDescriptionOfPWeights:= function( tbl, p, r )
    local blocks, i, ntbl, b, fus, bl, pos, infl;

    if not IsBound( r.blocks ) then
      blocks:= [];
      for i in [ 1 .. Length( r.normalizers ) ] do
        ntbl:= r.normalizers[i];
        if IsOrdinaryTable( ntbl ) then
          blocks[i]:= List( r.characters[i],
                            psi -> PRegularCorrespondent( ntbl, tbl, p, psi ) );
        else
          # 'ntbl' is a triple of character tables (N/K, M/K, M).
          # Compute the induced blocks of M/K.
          b:= List( r.characters[i],
                    psi -> PRegularCorrespondent( ntbl[1], ntbl[2], p, psi ) );

          # Identify irreducibles in the corresponding blocks of M.
          fus:= GetFusionMap( ntbl[3], ntbl[2] );
          if fus = fail then
            Error( "no fusion from <ntbl>[3] to <ntbl>[2] stored" );
          fi;
          bl:= PrimeBlocks( ntbl[2], p );
          pos:= List( b, x -> Position( bl.block, x ) );
          infl:= List( Irr( ntbl[2] ){ pos }, x -> x{ fus } );

          pos:= Position( Irr( ntbl[3] ), infl );

          # Compute the induced blocks of G.
          blocks[i]:= List( infl,
                        psi -> PRegularCorrespondent( ntbl[3], tbl, p, psi ) );
        fi;
      od;

      r.blocks:= blocks;
    fi;
end;


#############################################################################
##
#M  DescriptionOfPWeightsOp( <tbl>, <p> )
##
##  This is a method which accesses the local database.
##
InstallMethod( DescriptionOfPWeightsOp,
    [ "IsOrdinaryTable and IsLibraryCharacterTableRep", "IsPosInt" ],
    function( tbl, p )
    local id, data, ppart, size, pcore, tbls, irreds, i, currtbl, descr;

    id:= Identifier( tbl );

    data:= CTBlocks.Data( id );
    if data <> fail then
      # Fetch the data from the database.
      if IsBound( data.( p ) ) then
        data:= data.( p );
        if IsBound( data.radicalNormalizers ) then
          data:= data.radicalNormalizers;

          tbls:= [];
          irreds:= [];

          for i in [ 1 .. Length( data ) ] do
            if i = 1 then
              tbls[i]:= tbl;
              currtbl:= tbl;
            elif IsString( data[i].normalizer ) then
              tbls[i]:= CharacterTable( data[i].normalizer );
              currtbl:= tbls[i];
            else
              tbls[i]:= List( data[i].normalizer, CharacterTable );
              currtbl:= tbls[i][1];
            fi;

            ppart:= 1;
            size:= Size( currtbl );
            while size mod p = 0 do
              ppart:= ppart * p;
              size:= size / p;
            od;
            pcore:= ClassPositionsOfPCore( currtbl, p );
            ppart:= ppart / Sum( SizesConjugacyClasses( currtbl ){ pcore } );

            irreds[i]:= Filtered( Irr( currtbl ),
                chi -> chi[1] mod ppart = 0 and
                       IsSubset( ClassPositionsOfKernel( chi ), pcore ) );
          od;

          descr:= rec( normalizers:= tbls,
                       characters:= irreds );
          AddBlockDistributionToDescriptionOfPWeights( tbl, p, descr );
          return descr;

        fi;
      fi;
    fi;

    return fail;
    end );


#############################################################################
##
#M  DescriptionOfPWeightsOp( <tbl>, <p> )
##
InstallMethod( DescriptionOfPWeightsOp,
    [ "IsOrdinaryTable", "IsPosInt" ],
    function( tbl, p )
    local G, reps, tbls, irreds, i, currtbl, ppart, size, pcore, descr;

    if not HasUnderlyingGroup( tbl ) then
      return fail;
    fi;

    # Compute the radical p-subgroups and their normalizers.
    G:= UnderlyingGroup( tbl );
    reps:= RepresentativesRadicalPSubgroupsAndNormalizers( G, p );
    tbls:= [];
    irreds:= [];
    for i in [ 1 .. Length( reps ) ] do
      if Size( reps[i].normalizer ) = Size( tbl ) then
        currtbl:= tbl;
      else
        currtbl:= CharacterTable( reps[i].normalizer );
      fi;
      tbls[i]:= currtbl;
#T better compute p-core, switch to factor group,
#T compute its defect zero characters, and inflate them to the orig. table
#T -> so supply triple instead of character table!

      ppart:= 1;
      size:= Size( currtbl );
      while size mod p = 0 do
        ppart:= ppart * p;
        size:= size / p;
      od;
      pcore:= ClassPositionsOfPCore( currtbl, p );
      ppart:= ppart / Sum( SizesConjugacyClasses( currtbl ){ pcore } );

      irreds[i]:= Filtered( Irr( currtbl ),
          chi -> chi[1] mod ppart = 0 and
                 IsSubset( ClassPositionsOfKernel( chi ), pcore ) );
    od;

    descr:= rec( normalizers:= tbls,
                 characters:= irreds );
    AddBlockDistributionToDescriptionOfPWeights( tbl, p, descr );
    return descr;
    end );


#############################################################################
##
#M  DescriptionOfPWeightsOp( <tbl>, <p>, <b> )
##
InstallMethod( DescriptionOfPWeightsOp,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    local descr, result, i;

    descr:= DescriptionOfPWeights( tbl, p );
    if descr = fail then
      return fail;
    fi;
    AddBlockDistributionToDescriptionOfPWeights( tbl, p, descr );

    result:= rec( normalizers:= ShallowCopy( descr.normalizers ),
                  characters:= [] );
#T omit normalizers for which no characters??

    for i in [ 1 .. Length( descr.blocks ) ] do
      result.characters[i]:= descr.characters[i]{
                                 Positions( descr.blocks[i], b ) };
    od;

    return result;
    end );


#############################################################################
##
#M  ClassPositionsOfDefectClasses( <B> )
#M  ClassPositionsOfDefectClasses( <tbl>, <p>, <b> )
##
InstallMethod( ClassPositionsOfDefectClasses,
    [ "IsBlockOfCharacterTable" ],
    B -> ClassPositionsOfDefectClasses( UnderlyingCharacterTable( B ),
             UnderlyingCharacteristic( B ),
             NumberOfBlock( B ) ) );

InstallMethod( ClassPositionsOfDefectClasses,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    local blocks, pos, omega, coeffs, q, qp, orders, centralizers, i, frob,
          result, failed;

    blocks:= PrimeBlocks( tbl, p );
    pos:= Position( blocks.block, b );
    if pos = fail then
      return fail;
    fi;
    omega:= CentralCharacter( Irr( tbl )[ pos ] );
    coeffs:= CoefficientsOfOsimaIdempotent( tbl, p, b );
    q:= p ^ blocks.defect[b];
    qp:= q * p;
    orders:= OrdersClassRepresentatives( tbl );
    centralizers:= SizesCentralizers( tbl );

    # Search for classes of <p>-regular elements whose centralizer has
    # Sylow <p>-subgroup of order exactly `q'.
    result:= [];
    failed:= false;
    for i in [ 1 .. NrConjugacyClasses( tbl ) ] do

      if     orders[i] mod p <> 0
         and centralizers[i] mod q = 0
         and centralizers[i] mod qp <> 0 then

        # Check whether the reductions of `omega' and `coeffs' modulo
        # the chosen maximal ideal vanishes.
        frob:= FrobeniusCharacterValue( omega[i], p );
        if frob = fail then
          Info( InfoBlocks, 1,
                "ClassPositionsOfDefectClasses: ",
                "problem at class ", i );
          failed:= true;
          break;
        elif not IsZero( frob ) then
          frob:= FrobeniusCharacterValue( coeffs[i], p );
          if frob = fail then
            Info( InfoBlocks, 1,
                  "ClassPositionsOfDefectClasses: ",
                  "problem at class ", i );
            failed:= true;
            break;
          elif not IsZero( frob ) then
            Add( result, i );
          fi;
        fi;

      fi;

    od;

    if failed then
      Info( InfoBlocks, 1,
            "ClassPositionsOfDefectClasses: ",
            "cannot determine the set of defect classes" );
      return fail;
    else
      return result;
    fi;
    end );


#############################################################################
##
#M  ClassPositionsOfDefectClassesUpToGaloisConjugacy( <B> )
#M  ClassPositionsOfDefectClassesUpToGaloisConjugacy( <tbl>, <p>, <b> )
##
InstallMethod( ClassPositionsOfDefectClassesUpToGaloisConjugacy,
    [ "IsBlockOfCharacterTable" ],
    B -> ClassPositionsOfDefectClassesUpToGaloisConjugacy(
             UnderlyingCharacterTable( B ),
             UnderlyingCharacteristic( B ),
             NumberOfBlock( B ) ) );

InstallMethod( ClassPositionsOfDefectClassesUpToGaloisConjugacy,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    local blocks, omega, coeffs, q, qp, orders, centralizers, result, done,
          i, oo, orb, j, pow;

    blocks:= PrimeBlocks( tbl, p );
    if Length( blocks.block ) < b then
      return fail;
    fi;

    omega:= CentralCharacter( Irr( tbl )[ Position( blocks.block, b ) ] );
    # Note that the denominators of the coefficients of the idempotent
    # are coprime to 'p', so they can be ignored when one is interested
    # in whether or not the reduction modulo 'p' is zero.
    coeffs:= List( CoefficientsOfOsimaIdempotent( tbl, p, b ),
                   x -> x * DenominatorCyc( x ) );
    q:= p ^ blocks.defect[b];
    qp:= q * p;
    orders:= OrdersClassRepresentatives( tbl );
    centralizers:= SizesCentralizers( tbl );
    result:= [];

    # Consider Galois orbits of classes of <p>-regular elements
    # whose centralizer has Sylow <p>-subgroup of order exactly `q'.
    done:= BlistList( orders, [] );
    for i in [ 1 .. Length( orders ) ] do
      if not done[i] and
         orders[i] mod p <> 0 and
         centralizers[i] mod q = 0 and centralizers[i] mod qp <> 0 then
        # Compute the Galois orbit.
        oo:= orders[i];
        orb:= [ i ];
        for j in [ 2 .. oo-1 ] do
          if GcdInt( oo, j ) = 1 then
            pow:= PowerMap( tbl, j, i );
            AddSet( orb, pow );
            done[ pow ]:= true;
          fi;
        od;

        # Check whether all conjugates of `omega[i] * coeffs[i]'
        # are divisible by <p>; if no then one of the conjugates
        # is a defect class.
        if ForAny( orb, j -> not IsCycInt( omega[j] * coeffs[j] / p ) ) then
          UniteSet( result, orb );
        fi;
      fi;
    od;

    return result;
    end );


#############################################################################
##
#M  ClassPositionsOfDefectGroupOfClass( <tbl>, <p>, <c> )
##
InstallMethod( ClassPositionsOfDefectGroupOfClass,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, c )
    local orders, pow, roots, newroots, nccl, i, result, ppart, ord;

    orders:= OrdersClassRepresentatives( tbl );
    if Length( orders ) < c or orders[c] mod p = 0 then
      Error( "<c> must be the position of a <p>-regular class" );
    fi;

    pow:= PowerMap( tbl, p );
    roots:= [];
    newroots:= ClassOrbit( tbl, c );
    nccl:= Length( orders );
    while not IsEmpty( newroots ) do
#T better: run over the classes w.r.t. increasing order,
#T then one loop suffices!
      UniteSet( roots, newroots );
      newroots:= [];
      for i in [ 1 .. nccl ] do
        if ( not ( i in roots ) ) and pow[i] in roots
                                  and orders[i] <> orders[ pow[i] ] then
          AddSet( newroots, i );
        fi;
      od;
    od;

    result:= [];
    for i in roots do
      ppart:= 1;
      ord:= orders[i];
      while ord mod p = 0 do
        ppart:= ppart * p;
        ord:= ord / p;
      od;
      AddSet( result, PowerMap( tbl, ord, i ) );
    od;

    return result;
    end );


#############################################################################
##
#M  ClassPositionsOfDefectGroupOfBlock( <B> )
#M  ClassPositionsOfDefectGroupOfBlock( <tbl>, <p>, <b> )
##
InstallMethod( ClassPositionsOfDefectGroupOfBlock,
    [ "IsBlockOfCharacterTable" ],
    B -> ClassPositionsOfDefectGroupOfBlock( UnderlyingCharacterTable( B ),
             UnderlyingCharacteristic( B ),
             NumberOfBlock( B ) ) );

InstallMethod( ClassPositionsOfDefectGroupOfBlock,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    local pos;

    pos:= ClassPositionsOfDefectClassesUpToGaloisConjugacy( tbl, p, b );
    if pos = fail then
      return fail;
    fi;
    return ClassPositionsOfDefectGroupOfClass( tbl, p, pos[1] );
    end );


#############################################################################
##
#M  DefectGroup( <B> )
#M  DefectGroup( <tbl>, <p>, <b> )
##
InstallMethod( DefectGroup,
    [ "IsBlockOfCharacterTable" ],
    B -> DefectGroup( UnderlyingCharacterTable( B ),
             UnderlyingCharacteristic( B ),
             NumberOfBlock( B ) ) );

InstallMethod( DefectGroup,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )

    if not HasUnderlyingGroup( tbl ) then
      return fail;
    fi;

    b:= ClassPositionsOfDefectClassesUpToGaloisConjugacy( tbl, p, b );
    if b = fail then
      return fail;
    fi;

    return SylowSubgroup( Centralizer( UnderlyingGroup( tbl ),
               Representative( ConjugacyClasses( tbl )[ b[1] ] ) ), p );
    end );


#############################################################################
##
#M  NormalizerOfDefectGroup( <B> )
#M  NormalizerOfDefectGroup( <tbl>, <p>, <b> )
##
InstallMethod( NormalizerOfDefectGroup,
    [ "IsBlockOfCharacterTable" ],
    function( B )
    local p, N;

    p:= UnderlyingCharacteristic( B );
    N:= NormalizerOfDefectGroup( UnderlyingCharacterTable( B ),
            p,
            NumberOfBlock( B ) );
    if N <> fail then
      SetDefectGroup( B, PCore( N, p ) );
    fi;

    return N;
    end );

InstallMethod( NormalizerOfDefectGroup,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    local D, N;

    if not HasUnderlyingGroup( tbl ) then
      return fail;
    fi;

    D:= DefectGroup( tbl, p, b );
    if D = fail then
      return fail;
    fi;
    N:= Normalizer( UnderlyingGroup( tbl ), D );
    SetPCore( N, p, D );

    return N;
    end );


#############################################################################
##
#M  IsClassWithNormalDefectGroup( <tbl>, <p>, <c> )
##
InstallMethod( IsClassWithNormalDefectGroup,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, c )
    local orders, ppart, size;

    orders:= OrdersClassRepresentatives( tbl );
    if Length( orders ) < c or orders[c] mod p = 0 then
      Error( "<c> must be the position of a <p>-regular class in <tbl>" );
    fi;

    ppart:= 1;
    size:= SizesCentralizers( tbl )[c];
    while size mod p = 0 do
      ppart:= ppart * p;
      size:= size / p;
    od;
    if ppart = 1 then
      return true;
    fi;

    return Sum( SizesConjugacyClasses( tbl ){
                 ClassPositionsOfDefectGroupOfClass( tbl, p, c ) } ) = ppart;
    end );


#############################################################################
##
#M  IsBlockWithNormalDefectGroup( <tbl>, <p>, <b> )
##
InstallMethod( IsBlockWithNormalDefectGroup,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    local blocks;

    blocks:= PrimeBlocks( tbl, p );
    if Length( blocks.defect ) < b then
      Error( "<tbl> does not have a <b>-th <p>-block" );
    elif blocks.defect[b] = 0 then
      return true;
    elif ClassPositionsOfPCore( tbl, p ) = [ 1 ] then
      return false;
    fi;

    return Sum( SizesConjugacyClasses( tbl ){
                   ClassPositionsOfDefectGroupOfBlock( tbl, p, b ) } )
           = p^blocks.defect[b];
    end );


#############################################################################
##
#M  IsBlockWithNormalDefectGroup( <B> )
##
InstallMethod( IsBlockWithNormalDefectGroup,
    [ "IsBlockOfCharacterTable" ],
    B -> IsBlockWithNormalDefectGroup( UnderlyingCharacterTable( B ),
             UnderlyingCharacteristic( B ),
             NumberOfBlock( B ) ) );


#############################################################################
##
#M  IsClassWithCyclicDefectGroup( <tbl>, <p>, <c> )
##
InstallMethod( IsClassWithCyclicDefectGroup,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, c )
    local ord, ppart, size;

    ord:= OrdersClassRepresentatives( tbl );
    if Length( ord ) < c or ord[c] mod p = 0 then
      Error( "<c> must be the position of a <p>-regular class in <tbl>" );
    fi;

    ppart:= 1;
    size:= SizesCentralizers( tbl )[c];
    while size mod p = 0 do
      ppart:= ppart * p;
      size:= size / p;
    od;
    if ppart = 1 then
      return true;
    fi;

    return ppart in ord{ ClassPositionsOfDefectGroupOfClass( tbl, p, c ) };
    end );


#############################################################################
##
#M  IsBlockWithCyclicDefectGroup( <tbl>, <p>, <b> )
##
InstallMethod( IsBlockWithCyclicDefectGroup,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    local blocks;

    blocks:= PrimeBlocks( tbl, p );
    if Length( blocks.defect ) < b then
      Error( "<tbl> has no <b>-th <p>-block" );
    elif blocks.defect[b] = 0 then
      return true;
    fi;

    return p^blocks.defect[b] in OrdersClassRepresentatives( tbl ){
               ClassPositionsOfDefectGroupOfBlock( tbl, p, b ) };
    end );


#############################################################################
##
#M  IsBlockWithCyclicDefectGroup( <B> )
##
InstallMethod( IsBlockWithCyclicDefectGroup,
    [ "IsBlockOfCharacterTable" ],
    B -> IsBlockWithCyclicDefectGroup( UnderlyingCharacterTable( B ),
             UnderlyingCharacteristic( B ),
             NumberOfBlock( B ) ) );


#############################################################################
##
#M  TestBlockWithAbelianDefectGroup( <B> )
##
InstallMethod( TestBlockWithAbelianDefectGroup,
    [ "IsBlockOfCharacterTable" ],
    B -> TestBlockWithAbelianDefectGroup(
             UnderlyingCharacterTable( B ),
             UnderlyingCharacteristic( B ),
             NumberOfBlock( B ) ) );


#############################################################################
##
#M  TestBlockWithAbelianDefectGroup( <tbl>, <p>, <b> )
##
InstallMethod( TestBlockWithAbelianDefectGroup,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    local bl, ht, orders, cents, classes, ppart, dcl, result, pos, n, f, sum,
          sub, subbl, subresult;

    bl:= PrimeBlocks( tbl, p );
    ht:= Set( bl.height{ Positions( bl.block, b ) } ) = [ 0 ];

    orders:= OrdersClassRepresentatives( tbl );
    cents:= SizesCentralizers( tbl );
    classes:= SizesConjugacyClasses( tbl );
    ppart:= p^bl.defect[b];
    dcl:= ClassPositionsOfDefectGroupOfBlock( tbl, p, b );

    result:= rec( resultHZC:= ht,
                  result:= fail,
                  comment:= "no criterion yields a decision" );

    if not ht then

      # If not all ordinary irreducible characters in the block have
      # height zero then the defect group of the block is nonabelian,
      # by [KM13].
      result.result:= false;
      result.comment:= "not all irreducibles have height zero";

    elif bl.defect[b] <= 2 then

      # If the defect is at most 2 then the defect groups are abelian.
      result.result:= true;
      result.comment:= "defect at most 2";

    elif p^bl.defect[b] in orders{ dcl } then

      # The defect group D is cyclic.
      result.result:= true;
      result.comment:= "cyclic defect group";

    elif Set( orders{ dcl } ) = [ 1, 2 ] then

      # The defect group is an el. ab. 2-group.
      result.result:= true;
      result.comment:= "defect group is an el. ab. 2-group";

    elif IsPSolvableCharacterTable( tbl, p ) then

      # The conjecture is true for p-solvable groups.
      result.result:= true;
      result.comment:= "group is p-solvable";

    elif ForAny( dcl, i -> cents[i] mod ppart <> 0 ) then

      # If the defect groups contain elements whose centralizer order is
      # not divisible by the order of the defect group then the defect groups
      # are nonabelian.
      # (If we arrive here then we have a counterexample to Brauer's
      # height zero conjecture, so we get an error in the end.)
      result.result:= false;
      result.comment:= "defect group element with too small centralizer";

    elif ( Size( tbl ) / ppart ) mod p <> 0 then

      # For a block of maximal defect, the defect groups (the Sylow subgroups)
      # are abelian if and only if all ordinary irreducibles in the
      # principal p-block have height zero and the class lengths of all
      # p-elements are not divisible by p.
      pos:= Position( Irr( tbl ), TrivialCharacter( tbl ) );
      if bl.block[ pos ] = b then
        # We are asked about the principal block.
        result.result:= true;
      else
        # We take the result for the principal block.
        result.result:= TestBlockWithAbelianDefectGroup( tbl, p,
                            bl.block[ pos ] ).result;
      fi;
      result.comment:= "maximal defect";

    elif Sum( classes{ dcl } ) = ppart then

      # The defect group D is normal in G.
      # The height zero conjecture is true in this case,
      # by a theorem of Reynolds (Thm. 9.23 in [Nav98]).
      result.result:= true;
      result.comment:= "normal defect group";

    elif IsBlockWithAbelianDefectGroup( PrincipalBlock( tbl, p ) ) then

      # The Sylow p-subgroups are abelian.
      result.result:= true;
      result.comment:= "Sylow p-subgroups are abelian";

    elif HasUnderlyingGroup( tbl ) then

      # Compute the result from the group.
      result.result:= IsAbelian( DefectGroup( tbl, p, b ) );
      result.comment:= "computed from the defect group";

    fi;

    if result.resultHZC <> result.result and result.result <> fail then
      Error( "counterexample for Brauer's height zero conjecture for ",
             [ tbl, p, b ] );
    fi;

    return result;
    end );


#############################################################################
##
#M  IsBlockWithAbelianDefectGroup( <B> )
##
InstallMethod( IsBlockWithAbelianDefectGroup,
    [ "IsBlockOfCharacterTable" ],
    B -> TestBlockWithAbelianDefectGroup(
             UnderlyingCharacterTable( B ),
             UnderlyingCharacteristic( B ),
             NumberOfBlock( B ) ).result );


#############################################################################
##
#M  IsBlockWithAbelianDefectGroup( <tbl>, <p>, <b> )
##
InstallMethod( IsBlockWithAbelianDefectGroup,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    return TestBlockWithAbelianDefectGroup( tbl, p, b ).result;
    end );


#############################################################################
##
#F  CTBlocks.BrauerCorrespondent( <tblH>, <tblG>, <p>, <phi>,
#F                                                <d> )
##
##  <phi> is an irreducible character in the relevant <p>-block of <tblH>,
##  and <d> is either the <p>-defect of this block or `fail'.
##  (If <d> is known then we can discard all <p>-blocks of <tblG> of smaller
##  defect, but there are situations where we do not want to compute the
##  defect, for example if we do not know all irreducible characters of
##  <tblH>.)
##
CTBlocks.BrauerCorrespondent:=
    function( tblH, tblG, p, phi, d )
    local pos, fus, ind, blocks, orders, phivalues, i, omega;

    if IsIdenticalObj( tblH, tblG ) then
      # Compute the block number of 'phi'.
      pos:= Position( Irr( tblH ), phi );
      if pos = fail then
        Error( "<phi> is not an irreducible character of <tblH>" );
      fi;
      return PrimeBlocks( tblG, p ).block[ pos ];
    fi;

    fus:= FusionConjugacyClasses( tblH, tblG );
    if fus = fail then
      Error( "cannot determine the class fusion from <tblH> to <tblG>" );
    fi;

    # Compute \omega_{phi^G}.
    ind:= CentralCharacter( tblG, Induced( tblH, tblG, [ phi ], fus )[1] );
    blocks:= PrimeBlocks( tblG, p );
    orders:= OrdersClassRepresentatives( tblG );
    phivalues:= List( orders, Phi );

    # Identify the block B of G such that
    # \omega_{phi^G} - \omega_B is in the kernel of p-modular reduction.
    # If no such block exists then return `fail'.
    # Note that we have to check (representatives of Galois families of)
    # *all* classes, not only the p-regular classes,
    # and that ramification may occur at p-singular elements;
    # so we have to check a power of the difference.
    for i in [ 1 .. Length( blocks.defect ) ] do
      # If the induced block is defined then its defect is at least `d'.
      if d = fail or d <= blocks.defect[i] then
        omega:= CentralCharacter( Irr( tblG )[ Position( blocks.block, i ) ] );
        if ForAll( [ 1 .. Length( omega ) ],
             j -> IsCycInt( ( ind[j] - omega[j] )^phivalues[j] / p ) ) then
          return i;
        fi;
      fi;
    od;

    # Induction in the sense of Brauer is not defined.
    return fail;
    end;


#############################################################################
##
#M  BrauerCorrespondent( <tblH>, <tblG>, <p>, <b> )
##
InstallMethod( BrauerCorrespondent,
    [ "IsOrdinaryTable", "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tblH, tblG, p, b )
    local blocks, d, pos;

    blocks:= PrimeBlocks( tblH, p );
    d:= blocks.defect[b];
    pos:= Position( blocks.block, b );
    if pos = fail then
      Error( "<b> is not the position of a <p>-block in <tblH>" );
    elif IsIdenticalObj( tblH, tblG ) then
      return b;
    fi;

    return CTBlocks.BrauerCorrespondent( tblH, tblG, p,
                                        Irr( tblH )[ pos ], d );
    end );


#############################################################################
##
#M  BrauerCorrespondent( <B>, <tblG> )
##
InstallMethod( BrauerCorrespondent,
    [ "IsBlockOfCharacterTable", "IsOrdinaryTable" ],
    function( B, tblG )
    return BrauerCorrespondent( UnderlyingCharacterTable( B ), tblG,
             UnderlyingCharacteristic( B ),
             NumberOfBlock( B ) );
    end );


#############################################################################
##
#M  BrauerCorrespondent( <tblH>, <tblG>, <p>, <phi> )
##
InstallMethod( BrauerCorrespondent,
    [ "IsOrdinaryTable", "IsOrdinaryTable", "IsPosInt", "IsList" ],
    function( tblH, tblG, p, phi )
    return CTBlocks.BrauerCorrespondent( tblH, tblG, p, phi, fail );
    end );


#############################################################################
##
#M  PRegularCorrespondent( <tblH>, <tblG>, <p>, <phi> )
##
InstallMethod( PRegularCorrespondent,
    [ "IsOrdinaryTable", "IsOrdinaryTable", "IsPosInt", "IsList" ],
    function( tblH, tblG, p, phi )
    local pos, fus, ind, blocks, relevant, i;

    if IsIdenticalObj( tblH, tblG ) then
      # Compute the block number of 'phi'.
      pos:= Position( Irr( tblH ), phi );
      if pos = fail then
        Error( "<phi> is not an irreducible character of <tblH>" );
      fi;
      return PrimeBlocks( tblG, p ).block[ pos ];
    fi;

    fus:= FusionConjugacyClasses( tblH, tblG );
    if fus = fail then
      Error( "cannot determine the class fusion from <tblH> to <tblG>" );
    fi;

    ind:= CentralCharacter( tblG, Induced( tblH, tblG, [ phi ], fus )[1] );
    blocks:= PrimeBlocks( tblG, p );
    relevant:= ShallowCopy( blocks.relevant );
    for i in [ 1 .. Length( blocks.defect ) ] do
      if SameBlock( p, ind,
             CentralCharacter( Irr( tblG )[ Position( blocks.block, i ) ] ),
             relevant ) then
        return i;
      fi;
    od;

    # <p>-regular induction is not defined.
    return fail;
    end );


#############################################################################
##
#M  PRegularCorrespondent( <tblH>, <tblG>, <p>, <b> )
##
InstallMethod( PRegularCorrespondent,
    [ "IsOrdinaryTable", "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tblH, tblG, p, b )
    local blocks, pos;

    blocks:= PrimeBlocks( tblH, p );
    pos:= Position( blocks.block, b );
    if pos = fail then
      Error( "<b> is not the position of a <p>-block in <tblH>" );
    elif IsIdenticalObj( tblH, tblG ) then
      return b;
    fi;

    return PRegularCorrespondent( tblH, tblG, p, Irr( tblH )[ pos ] );
    end );


#############################################################################
##
#M  PRegularCorrespondent( <B>, <tblG> )
##
InstallMethod( PRegularCorrespondent,
    [ "IsBlockOfCharacterTable", "IsOrdinaryTable" ],
    function( B, tblG )
    return PRegularCorrespondent( UnderlyingCharacterTable( B ), tblG,
             UnderlyingCharacteristic( B ),
             NumberOfBlock( B ) );
    end );


#############################################################################
##
#F  ConstituentsInRestriction( <tblH>, <tblG>, <p>, <b> )
##
##  The implementation follows the algorithm underlying [EH92, Theorem 1.1].
##
InstallGlobalFunction( ConstituentsInRestriction,
    function( tblH, tblG, p, b )
    local H, G, D, L, beta, tblL, blH, defect, blL, i, Dclasses, irr, chi,
          K, indchi, tblK, blK, constit;

    if not HasUnderlyingGroup( tblH ) then
      Error( "<tblH> must store an underlying group");
    elif not HasUnderlyingGroup( tblG ) then
      Error( "<tblG> must store an underlying group");
    fi;

    # Let $B'$ be the `b'-th block of $H$.
    # Compute the defect group $D$ of $B'$.
    H:= UnderlyingGroup( tblH );
    G:= UnderlyingGroup( tblG );
    D:= DefectGroup( tblH, p, b );

    # Let $L = D \cdot C_H(D)$.
    L:= ClosureGroup( D, Centralizer( H, D ) );

    # Let $\beta'$ be the unique block of $L$ such that $(\beta')^H = B'$.
    # Note that there is a unique block $b'$ of $N_H(D)$
    # such that $(b')^H = B'$, and that $b'$ is in fact regular relative to
    # $L$ (Lemma~5.5.14 in Nagao-Tsushima), so $b'$ has defect group $D$.
    # Now $L$ is normal in $N_H(D)$, and $\beta'$ is covered by $b'$;
    # By Exercise~5.5.18 in Nagao-Tsushima,
    # also $\beta'$ has defect group $D$,
    # thus we conclude that the situation is admissible.
    # (Hence it suffices to deal with Brauer correspondents.)
    beta:= 0;
    tblL:= CharacterTable( L );
    blH:= PrimeBlocks( tblH, p );
    defect:= blH.defect[b];
    blL:= PrimeBlocks( tblL, p );
    for i in [ 1 .. Length( blL.defect ) ] do
      if blL.defect[i] = defect then
        if BrauerCorrespondent( tblL, tblH, p, i ) = b then
          beta:= i;
          break;
        fi;
      fi;
    od;
    Dclasses:= ClassPositionsOfNormalSubgroup( tblL, D );

    # Let $\chi$ be the unique irreducible character in $\beta'$
    # that has $D$ in its kernel.
    # (This character is called the canonical character of $\beta'$,
    # see 5.8.12 in Nagao-Tsushima.)
    irr:= Irr( tblL );
    chi:= First( [ 1 .. NrConjugacyClasses( tblL )] ,
              x -> blL.block[x] = i and
   IsSubset( ClassPositionsOfKernel( irr[x] ), Dclasses ) );
    chi:= irr[ chi ];

    # Form $\chi^K$ where $K = D \cdot C_G(D)$.
    K:= ClosureGroup( D, Centralizer( G, D ) );
    indchi:= InducedClassFunction( chi, K );

    # Note:
    # 1. If $\hat{\chi}$ is the character of $L/D$ corresponding to
    #    $\chi \in Irr(L)$ then $\hat{\chi}^{K/D}$ is the character of $K/D$
    #    corresponding to $\chi^K$;
    #    This means that we may work with $\chi$ instead of $\hat{\chi}$,
    #    although it might be worthwhile also for GAP to compute
    #    in the factor group by $D$.
    # 2. The blocks of $K$ are in bijection with the blocks of $K/D$,
    #    by Theorem~5.8.11 in Nagao-Tsushima, since $|K/C_K(D)$ is a power
    #    of $p$.
    #    This means that we may work with $\chi^K$ instead of
    #    $\hat{\chi}^{K/D}$,

    # $\chi^K$ decomposes into constituents in the blocks
    # $\beta_1, \beta_2, \ldots, \beta_r$ of $K$.
    K:= ClosureGroup( D, Centralizer( G, D ) );
    tblK:= CharacterTable( K );
    blK:= PrimeBlocks( tblK, p );
    constit:= [];
    irr:= Irr( tblK );
    Dclasses:= ClassPositionsOfNormalSubgroup( tblK, D );
    for i in [ 1 .. Length( blK.block ) ] do
      if     not blK.block[i] in constit
         and IsSubset( ClassPositionsOfKernel( irr[i] ), Dclasses )
         and ScalarProduct( tblK,  indchi, irr[i] ) <> 0 then
        Add( constit, blK.block[i] );
      fi;
    od;

    # Then $\beta_1^G, \beta_2^G, \ldots, \beta_r^G$ are exactly
    # the blocks $B$ of $G$ with $B' \mid B_{H \times H}$.
    return Set( List( constit,
                      beta -> BrauerCorrespondent( tblK, tblG, p, beta ) ) );
    end );


#############################################################################
##
#M  AlperinBurryCorrespondent( <tblH>, <tblG>, <p>, <b> )
##
InstallMethod( AlperinBurryCorrespondent,
    [ "IsOrdinaryTable", "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tblH, tblG, p, b )
#T in case of admissibility, we may simply take the Brauer correspondent
#T (cheaper?)
    b:= ConstituentsInRestriction( tblH, tblG, p, b );
    if Length( b ) = 1 then
      return b[1];
    else
      return fail;
    fi;
    end );


#############################################################################
##
#M  AlperinBurryCorrespondent( <B>, <tblG> )
##
InstallMethod( AlperinBurryCorrespondent,
    [ "IsBlockOfCharacterTable", "IsOrdinaryTable" ],
    function( B, tblG )
    return AlperinBurryCorrespondent( UnderlyingCharacterTable( B ), tblG,
             UnderlyingCharacteristic( B ),
             NumberOfBlock( B ) );
    end );


#############################################################################
##
#M  IsAdmissibleBlock( <B>, <tblG> )
##
InstallMethod( IsAdmissibleBlock,
    [ "IsBlockOfCharacterTable", "IsOrdinaryTable" ],
    function( B, tblG )
    return IsAdmissibleBlock( UnderlyingCharacterTable( B ), tblG,
             UnderlyingCharacteristic( B ),
             NumberOfBlock( B ) );
    end );


#############################################################################
##
#M  IsAdmissibleBlock( <tblH>, <tblG>, <p>, <b> )
##
InstallMethod( IsAdmissibleBlock,
    [ "IsOrdinaryTable", "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tblH, tblG, p, b )
    local H, G;

    if not HasUnderlyingGroup( tblH ) or not HasUnderlyingGroup( tblG ) then
      return fail;
    fi;

    H:= UnderlyingGroup( tblH );
    G:= UnderlyingGroup( tblG );

    return IsSubset( H, Centralizer( G, DefectGroup( tblH, p, b ) ) );
    end );


#############################################################################
##
#F  IsMultiplicityOne( <tblH>, <tblG>, <p>, <b> )
##
##  Let <tblH> and <tblG> be the ordinary character tables of two groups $H$
##  and $G$, respectively, with $H \leq G$, <p> a prime integer,
##  and <b> a positive integer denoting the number of a <p>-block of <tblH>.
##  `IsMultiplicityOne' returns `true' if the multiplicity of the <b>-th
##  block of <tblH> as a direct summand in the restriction of $RG$ to <tblH>,
##  viewed as an $H \times H$ module, is proved to be equal to $1$;
##  if the multiplicity is proved to be different from $1$ then `false' is
##  returned,
##  and if the criteria used do not suffice to decide the question
##  then `fail' is returned.
##
InstallGlobalFunction( IsMultiplicityOne, function( tblH, tblG, p, b )
    local H, G, blH, ind, D, N, M, bprime, tblN, defect, blN, i, pi, Kbar,
          Lbar, Dclasses, irr, betaprime;

    if not HasUnderlyingGroup( tblH ) or not HasUnderlyingGroup( tblG ) then
      return fail;
    fi;

    H:= UnderlyingGroup( tblH );
    G:= UnderlyingGroup( tblG );

    if IsAdmissibleBlock( tblH, tblG, p, b ) then

      # If the <b>-th block is admissible in $G$ then it occurs with
      # multiplicity one (Theorem~5.10.12 in Nagao-Tsushima).
      return true;

    elif not IsSubset( H, Centralizer( G, H ) ) then

      # Multiplicity one implies $C_G(H) \leq H$,
      # by the usual double coset argument.
      return false;

    fi;

    # If the <b>-th block of <tblH> has defect zero then
    # the multiplicity is one if and only if the unique character in
    # this block induces irreducibly to $G$.
    blH:= PrimeBlocks( tblH, p );
    if blH.defect[b] = 0 then
      ind:= Irr( tblH )[ Position( blH.block, b ) ]^tblG;
      return ScalarProduct( ind, ind ) = 1;
    fi;

    # Let $D$ be the defect group of the <b>-th block $B'$ of <tblH>,
    # and let $b'$ be the unique block of $N = N_H(D)$ corresponding to $B'$.
    # By Proposition~5.4 in the paper of Ellers/Hill, the multiplicity of
    # $B'$ in $RG_{H \times H}$ equals the multiplicity of $b'$ in
    # $RM_{N \times N}$, where $M = N_G(D)$.
    D:= DefectGroup( tblH, p, b );
    if not IsNormal( G, D ) then
      N:= Normalizer( H, D );
      M:= Normalizer( G, D );
      bprime:= 0;
      tblN:= CharacterTable( N );
      defect:= blH.defect[b];
      blN:= PrimeBlocks( tblN, p );
      for i in [ 1 .. Length( blN.defect ) ] do
        if blN.defect[i] = defect then
          if BrauerCorrespondent( tblN, tblH, p, i ) = b then
            bprime:= i;
            break;
          fi;
        fi;
      od;
      return IsMultiplicityOne( CharacterTable( N ), CharacterTable( M ), p,
                                bprime );
#T really recursion?
    fi;

    # Now $H = N$ and $G = M$.
    if IsCentral( G, D ) then

      # If $N_G(D) = C_G(D)$ and $N_H(D) = C_H(D)$ hold then
      # Theorem~4.2 of Ellers/Hill can be applied,
      # that is, we can factor out $D$ without changing multiplicities.
      # (In the new situation, the defect is zero.)
      pi:= NaturalHomomorphismByNormalSubgroup( G, D );
      Kbar:= Image( pi, G );
      Lbar:= Image( pi, H );
      Dclasses:= ClassPositionsOfNormalSubgroup( tblH, D );
      irr:= Irr( tblH );
      betaprime:= First( [ 1 .. NrConjugacyClasses( tblH )] ,
              x -> blH.block[x] = b and
                   IsSubset( ClassPositionsOfKernel( irr[x] ), Dclasses ) );

      return IsMultiplicityOne( CharacterTable( Lbar ),
                                CharacterTable( Kbar ),
                                p, betaprime );

    fi;

    # There are currently no further criteria, we give up.
    return fail;
end );


#############################################################################
##
#F  MultiplicityOfBlockInRestriction( <tblH>, <tblG>, <p>, <b>, <B> )
##
##  The implementation uses Theorem~7.1 in~\cite{EH92}.
##  That is, it is assumed that $G = C_G(D) D$, and $D \leq H$ is the defect
##  group of the <b>-th block of $H$;
##  There is a unique character $\psi$ in the block $b$ that does not have
##  $D$ in its kernel (the canonical character of the block),
##  and if $\psi^G = \sum_{\chi\in B} n_{\chi} \chi + \theta$ where $\theta$
##  has no irreducible constituents in $B$ then the required multiplicity is
##  $\sum_{\chi \in B} n_{\chi}^2$.
##
InstallGlobalFunction( MultiplicityOfBlockInRestriction,
    function( tblH, tblG, p, b, B )
    local G, D, Dclasses, blH, irrH, i, psi, ind, blG, irrG, mult;

    G:= UnderlyingGroup( tblG );
    D:= DefectGroup( tblH, p, b );
    if not IsNormal( G, D )
       or not IsSubset( ClosureGroup( D, Centralizer( G, D ) ), G ) then
      return fail;
    fi;

    # Find the canonical character $\psi$.
    Dclasses:= ClassPositionsOfNormalSubgroup( tblH, D );
    blH:= PrimeBlocks( tblH, p );
    irrH:= Irr( tblH );
    psi:= irrH[ First( [ 1 .. Length( blH.block ) ],
                       i -> blH.block[i] = b
                           and IsSubset( ClassPositionsOfKernel( irrH[i] ),
                                         Dclasses ) ) ];

    # Compute the scalar products.
    ind:= InducedClassFunction( psi, G );
    blG:= PrimeBlocks( tblG, p );
    irrG:= Irr( tblG );
    mult:= 0;
    for i in [ 1 .. Length( irrG ) ] do
      if blG.block[i] = B then
        mult:= mult + ScalarProduct( ind, irrG[i] )^2;
      fi;
    od;

    # Return the result.
    return mult;
end );


#############################################################################
##
#M  WheelerCorrespondent( <tblH>, <tblG>, <p>, <b> )
##
##  We use the characterization from [Whe94, Proposition 1.4].
##  Namely, $b^{extG}$ is defined in the sense of Wheeler and $b^{extG} = B$
##  holds if and only if for any character $\Phi$ in $b$,
##  $B$ is the unique block of $G$ with the property
##  $(\Phi^B(1) / \Phi^G(1))^* \noteq 0$,
##  where $\Phi^B$ is the restriction of $\Phi^G$ to $B$.
##
InstallMethod( WheelerCorrespondent,
    [ "IsOrdinaryTable", "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tblH, tblG, p, b )
    local pos, fus,
          blH,     #`p'-blocks of `tblH'
          Phi,     # one character in the `b'-th block
          ind,     # induced character of `Phi'
          blG,     # `p'-blocks of `tblG'
          irrG,    # list of irreducible characters of $G$
          m,       # list of scalar products
          constit, # positions of constituents
          bl,      # block positions of the constituents
          degrees, # irreducible degrees of $G$
          mult,    # exponent of the `p'-part of `ind'
          B,       # position of the induced block
          inv,     # inverse map of the block distribution
          i,       # loop over `bl'
          j,       # loop over `inv'
          psi;     # degree of the projection to the `i'-th block

    if IsIdenticalObj( tblH, tblG ) then
      return b;
    fi;

    # Induce a character $\Phi$ in the `b'-th block to $G$.
    fus:= FusionConjugacyClasses( tblH, tblG );
    if fus = fail then
      Error( "cannot determine the class fusion from <tblH> to <tblG>" );
    fi;

    blH:= PrimeBlocks( tblH, p );
    Phi:= Irr( tblH )[ First( [ 1 .. Length( blH.block ) ],
                      i -> blH.block[i] = b ) ];
    ind:= Induced( tblH, tblG, [ Phi ], fus );

    # If all constituents of $\Phi^G$ lie in the same block of $G$
    # then this is the induced block in Wheeler's sense.
    blG:= PrimeBlocks( tblG, p );
    irrG:= Irr( tblG );
    m:= MatScalarProducts( tblG, irrG, ind )[1];
    constit:= Filtered( [ 1 .. Length( m ) ], x -> m[x] <> 0 );
    bl:= Set( blG.block{ constit } );
    if Length( bl ) = 1 then
      return bl[1];
    fi;

    # Let `mult' be the exponent of the `p'-part of the degree of $\Phi^G$.
    # If there is exactly one `p'-block $B$ of $G$ such that the degree of
    # the projection of $\Phi^G$ to $B$ has the same `p'-part as $\Phi^G$
    # then $B$ is the induced block,
    # otherwise extended induction is not defined.
    degrees:= List( irrG, x -> x[1] );
    mult:= Number( Factors( ind[1][1] ), x -> x = p );
    B:= 0;
    inv:= InverseMap( blG.block );
    for i in bl do
      j:= inv[i];
      if IsList(j) then
        psi:= m{j}*degrees{j};
      else
        psi:= m[j]*degrees[j];
      fi;
      if Number( Factors( psi ), x -> x = p ) = mult then
        if B = 0 then
          B:= i;
        else
          return fail;
        fi;
      fi;
    od;
    if B = 0 then
      return fail;
    else
      return B;
    fi;
    end );


#############################################################################
##
#M  WheelerCorrespondent( <B>, <tblG> )
##
InstallMethod( WheelerCorrespondent,
    [ "IsBlockOfCharacterTable", "IsOrdinaryTable" ],
    function( B, tblG )
    return WheelerCorrespondent( UnderlyingCharacterTable( B ), tblG,
             UnderlyingCharacteristic( B ),
             NumberOfBlock( B ) );
    end );


#############################################################################
##
#M  WheelerCorrespondent( <tblH>, <tblG>, <p>, <phi> )
##
InstallMethod( WheelerCorrespondent,
    [ "IsOrdinaryTable", "IsOrdinaryTable", "IsPosInt", "IsList" ],
    function( tblH, tblG, p, phi )
    local pos, b;

    # Compute the block number of 'phi'.
    pos:= Position( Irr( tblH ), phi );
    if pos = fail then
      Error( "<phi> is not an irreducible character of <tblH>" );
    fi;
    b:= PrimeBlocks( tblH, p ).block[ pos ];

    if IsIdenticalObj( tblH, tblG ) then
      return b;
    else
      return WheelerCorrespondent( tblH, tblG, p, b );
    fi;
    end );


#############################################################################
##
#M  IsRealClass( <tbl>, <c> )
##
InstallMethod( IsRealClass,
    [ "IsOrdinaryTable", "IsPosInt" ],
    function( tbl, c )
    if c = 1 then
      return true;
    elif HasInverseClasses( tbl ) then
      return InverseClasses( tbl )[c] = c;
    elif HasIrr( tbl ) then
      return ForAll( Irr( tbl ), chi -> chi[c] = GaloisCyc( chi[c], -1 ) );
    else
      return PowerMap( tbl, -1, c ) = c;
    fi;
    end );


#############################################################################
##
#M  IsStronglyRealClass( <tbl>, <c> )
##
##  For a real class, we check whether it consists of products of at most two
##  involutions.
##  Since the class multiplication coefficients <M>c_{ijc}</M> and
##  <M>c_{jic}</M> are equal, we have to check only <M>i \leq j</M>.
##
InstallMethod( IsStronglyRealClass,
    [ "IsOrdinaryTable", "IsPosInt" ],
    function( tbl, c )
    local orders, inv, j, invj, i;

    orders:= OrdersClassRepresentatives( tbl );
    if orders[c] <= 2 then
      return true;
    elif not IsRealClass( tbl, c ) then
      return false;
    else
      inv:= Positions( orders, 2 );
      for j in [ 1 .. Length( inv ) ] do
        invj:= inv[j];
        for i in [ 1 .. j ] do
          if ClassMultiplicationCoefficient( tbl, inv[i], invj, c ) <> 0 then
            return true;
          fi;
        od;
      od;
      return false;
    fi;
    end );


#############################################################################
##
#M  IsRealBlock( <B> )
#M  IsRealBlock( <tbl>, <p>, <b> )
##
InstallMethod( IsRealBlock,
    [ "IsBlockOfCharacterTable" ],
    function( B )
    local irr;

    irr:= Irr( B );
    return ComplexConjugate( irr[1] ) in irr;
    end );

InstallMethod( IsRealBlock,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    local irr, pos;

    irr:= Irr( tbl );
    pos:= Positions( PrimeBlocks( tbl, p ).block, b );

    if IsEmpty( pos ) then
      return fail;
    fi;
    return Position( irr, ComplexConjugate( irr[ pos[1] ] ) ) in pos;
    end );


#############################################################################
##
#M  IsStronglyRealBlock( <B> )
#M  IsStronglyRealBlock( <tbl>, <p>, <b> )
##
InstallMethod( IsStronglyRealBlock,
    [ "IsBlockOfCharacterTable" ],
    function( B )
    local tbl, proxy, c;

    if IsRealBlock( B ) = true then
      # Check the existence of a strongly real defect class.
      tbl:= UnderlyingCharacterTable( B );
      proxy:= GaloisData( tbl ).proxy;
      for c in ClassPositionsOfDefectClassesUpToGaloisConjugacy( B ) do
        # It is sufficient to check only Galois representatives.
        if proxy[c] = c and IsStronglyRealClass( tbl, c ) then
          return true;
        fi;
      od;
      return false;
    else
      return false;
    fi;
    end );

InstallMethod( IsStronglyRealBlock,
    [ "IsOrdinaryTable", "IsPosInt", "IsPosInt" ],
    function( tbl, p, b )
    local real, proxy, c;

    real:= IsRealBlock( tbl, p, b );
    if real = true then
      # Check the existence of a strongly real defect class.
      proxy:= GaloisData( tbl ).proxy;
      for c in ClassPositionsOfDefectClassesUpToGaloisConjugacy( tbl, p, b ) do
        # It is sufficient to check only Galois representatives.
        if proxy[c] = c and IsStronglyRealClass( tbl, c ) then
          return true;
        fi;
      od;
      return false;
    else
      # 'real' can be 'false' or 'fail'.
      return real;
    fi;
    end );


#############################################################################
##
##  Section "Selecting blocks according to their invariants"
##


#############################################################################
##
#V  CTBlocks.BlockInvariantsColumns
##
##  <#GAPDoc Label="CTBlocks.BlockInvariantsColumns">
##  <ManSection>
##  <Heading>Customizing the columns of block invariants overviews</Heading>
##  <Var Name="CTBlocks.BlockInvariantsColumns"/>
##
##  <Description>
##  This variable defines the columns of the overviews that are computed by
##  <Ref Func="BlockInvariants"/>, <Ref Func="DisplayBlockInvariants"/>, and
##  <Ref Func="BrowseBlockInvariants"/>.
##  Each column is described by a record with the following components.
##  <P/>
##  <List>
##  <Mark><C>label</C></Mark>
##  <Item>
##    a string that serves as a column label in the table header,
##  </Item>
##  <Mark><C>fun</C></Mark>
##  <Item>
##    a unary function that takes a block object as its argument,
##    for example <Ref Attr="Defect" Label="for a block object"/>,
##  </Item>
##  <Mark><C>align</C></Mark>
##  <Item>
##    one of <C>"left"</C> or <C>"right"</C>, meaning the alignment of the
##    values in the printed table,
##  </Item>
##  <Mark><C>replace</C> (optional)</Mark>
##  <Item>
##    if present, this must be a list of length three,
##    the first entry being the list of admissible values,
##    the second entry being the list of the corresponding strings
##    that shall be printed in the overview,
##    and the third entry being the list of the corresponding strings
##    that shall be shown in the table header when the output is restricted
##    to the given value;
##    the same strings are also shown in the category rows of the
##    browse table shown by <Ref Func="BrowseBlockInvariants"/>.
##    If no component <C>replace</C> is present then
##    <Ref Attr="String" BookName="ref"/> is applied to the invariant in
##    question, and the label together with the value will be used
##    in the header and in category rows of the printed table.
##  </Item>
##  </List>
##  <P/>
##  <Example>
##  gap> CTBlocks.BlockInvariantsColumns[1];
##  rec( align := "left", fun := function( B ) ... end, label := "G" )
##  </Example>
##  <P/>
##  The columns of the overview tables can be customized via the optional
##  record that can be entered as the last argument to
##  <Ref Func="BlockInvariants"/>, <Ref Func="DisplayBlockInvariants"/>, and
##  <Ref Func="BrowseBlockInvariants"/>.
##  If this record contains the component <C>columns</C> with value a list
##  of records as described above then this list of columns is considered
##  instead of the default list <Ref Var="CTBlocks.BlockInvariantsColumns"/>.
##  This way one can reduce, extend, or reorder the columns that are shown.
##  <P/>
##  <Example>
##  gap> cols:= List( [ "c", "a", "n", "f" ],
##  >                 l -> First( CTBlocks.BlockInvariantsColumns,
##  >                             r -> r.label = l ) );;
##  gap> t:= CharacterTable( "2.A5" );; 
##  gap> DisplayBlockInvariants( t, rec( columns:= cols ) );
##  Block invariants for 2.A5
##  
##  -----------------
##  | c | a | n | f |
##  -----------------
##  | - | - | - | + |
##  | + | + | + | + |
##  | + | + | - | - |
##  | + | + | + | - |
##  | + | + | + | - |
##  | + | + | - | + |
##  | + | + | + | + |
##  | + | + | - | - |
##  | + | + | + | - |
##  | + | + | - | + |
##  -----------------
##  </Example>
##  <P/>
##  By default, the following columns appear.
##  <P/>
##  <List>
##  <Mark><C>G</C> (omitted if only one character table is given)</Mark>
##  <Item>
##    the identifier of the underlying character table,
##  </Item>
##  <Mark><C>b</C></Mark>
##  <Item>
##    the block number,
##  </Item>
##  <Mark><C>d</C></Mark>
##  <Item>
##    the defect of the block,
##  </Item>
##  <Mark><C>k</C></Mark>
##  <Item>
##    the number of ordinary irreducibles in the block,
##  </Item>
##  <Mark><C>l</C></Mark>
##  <Item>
##    the number of modular irreducibles in the block,
##  </Item>
##  <Mark><C>c</C></Mark>
##  <Item>
##    <C>+</C> if the defect groups of the block are cyclic,
##    <C>-</C> otherwise,
##  </Item>
##  <Mark><C>a</C></Mark>
##  <Item>
##    <C>+</C> if the defect groups of the block are abelian,
##    <C>-</C> if they are nonabelian,
##    <C>?</C> if the character table does not determine this property,
##  </Item>
##  <Mark><C>n</C></Mark>
##  <Item>
##    <C>+</C> if the defect groups of the block are normal,
##    <C>-</C> otherwise,
##  </Item>
##  <Mark><C>f</C></Mark>
##  <Item>
##    <C>+</C> if the block is faithful,
##    <C>-</C> otherwise,
##  </Item>
##  <Mark><C>r</C></Mark>
##  <Item>
##    <C>+</C> if the block is real,
##    <C>-</C> otherwise,
##  </Item>
##  <Mark><C>sr</C></Mark>
##  <Item>
##    <C>+</C> if the block is strongly real,
##    <C>-</C> otherwise,
##  </Item>
##  </List>
##  <P/>
##  <Example>
##  gap> List( CTBlocks.BlockInvariantsColumns, r -> r.label );
##  [ "G", "p", "b", "d", "k", "l", "c", "a", "n", "f", "r", "sr" ]
##  </Example>
##  <P/>
##  Another supported component of the optional record is <C>header</C>.
##  If it is bound then its value must be a list of strings which apper
##  as the <C>header</C> component of the record returned by
##  <Ref Func="BlockInvariants"/>
##  and which are shown as a header above the tables printed by
##  <Ref Func="DisplayBlockInvariants"/> and
##  <Ref Func="BrowseBlockInvariants"/>.
##  The default for the header is <C>[ "Block invariants" ]</C> except that
##  the line is extended by the identifier of the character table if exactly
##  one character table is entered as the first argument.
##  <P/>
##  <Example>
##  gap> DisplayBlockInvariants( t, Defect, IsPosInt,
##  >      rec( columns:= cols,
##  >           header:= [ "only Boolean invariants" ] ) );
##  only Boolean invariants
##  
##  -----------------
##  | c | a | n | f |
##  -----------------
##  | - | - | - | + |
##  | + | + | + | + |
##  | + | + | - | - |
##  | + | + | - | + |
##  | + | + | - | - |
##  | + | + | - | + |
##  -----------------
##  </Example>
##  <P/>
##  (The user preference <C>DisplayFunction</C> from the
##  <Package>AtlasRep</Package> had been changed before showing examples
##  for <Ref Func="DisplayBlockInvariants"/>.
##  Now we reset this preference to its original value.)
##  <P/>
##  <Example>
##  gap> SetUserPreference( "AtlasRep", "DisplayFunction", origpref );
##  </Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
CTBlocks.BlockInvariantsColumns:= [
  rec( label:= "G",
       fun:= B -> Identifier( UnderlyingCharacterTable( B ) ),
       align:= "left" ),
  rec( label:= "p",
       fun:= UnderlyingCharacteristic,
       align:= "right" ),
  rec( label:= "b",
       fun:= NumberOfBlock,
       align:= "right" ),
  rec( label:= "d",
       fun:= Defect,
       align:= "right" ),
  rec( label:= "k",
       fun:= B -> DimensionsOfDecompositionMatrix( B )[1],
       align:= "right" ),
  rec( label:= "l",
       fun:= B -> DimensionsOfDecompositionMatrix( B )[2],
       align:= "right" ),
  rec( label:= "c",
       fun:= IsBlockWithCyclicDefectGroup,
       align:= "right",
       replace:= [ [ true, false ],
                   [ "+", "-" ],
                   [ "cyclic defect group", "noncyclic defect group" ] ] ),
  rec( label:= "a",
       fun:= IsBlockWithAbelianDefectGroup,
       align:= "right",
       replace:= [ [ true, false, fail ],
                   [ "+", "-", "?" ],
                   [ "abelian defect group", "nonabelian defect group",
                     "unknown whether defect group is abelian" ] ] ),
  rec( label:= "n",
       fun:= IsBlockWithNormalDefectGroup,
       align:= "right",
       replace:= [ [ true, false ],
                   [ "+", "-" ],
                   [ "normal defect group", "nonnormal defect group" ] ] ),
  rec( label:= "f",
       fun:= IsFaithful,
       align:= "right",
       replace:= [ [ true, false ],
                   [ "+", "-" ],
                   [ "faithful", "nonfaithful" ] ] ),
  rec( label:= "r",
       fun:= IsRealBlock,
       align:= "right",
       replace:= [ [ true, false ],
                   [ "+", "-" ],
                   [ "real", "nonreal" ] ] ),
  rec( label:= "sr",
       fun:= IsStronglyRealBlock,
       align:= "right",
       replace:= [ [ true, false ],
                   [ "+", "-" ],
                   [ "strongly real", "not strongly real" ] ] ),
  ];


#############################################################################
##
#F  CTBlocks.AccessFunction( <name>, <args>, <mode> )
##
##  This function is used by 'AllPBlocks', 'OnePBlock', 'BlockInvariants',
##  'DisplayBlockInvariants', and 'BrowseBlockInvariants'.
##
##  <args> is the list of arguments of these functions.
##  <name> is used in error messages, it should be the name of the calling
##  function.
##  <mode> must be one of '"all"', '"one"', '"inv"', '"str"'.
##
##  Depending on <mode>, the function returns
##  - '"all"':
##    all blocks that satisfy the conditions in <args>,
##  - '"one"':
##    one block that satisfies the conditions in <args>
##    if such a block exists, and 'fail' otherwise,
##  - '"inv"':
##    a record with the components 'header' (a list of strings)
##    and 'list' (the matrix of invariants of the blocks that satisfy the
##    conditions described by <args>; the columns of the matrix are a subset
##    of 'CTBlocks.BlockInvariantsColumns',
##    perhaps customized via the last entry in <args>),
##  - '"str"':
##    a record as in the previous case, where the matrix in the 'list'
##    component is obtained by replacing the invariants by suitable strings,
##    as described by the records in 'CTBlocks.BlockInvariantsColumns'.
##
CTBlocks.AccessFunction:= function( name, args, mode )
    local tbls, blocks, result, arec, columns, replace, char, cond, i, fun,
          vals, r, pos, tbl, primes, p, B;

    if Length( args ) = 0 then
      Error( "usage: ", name, "(<tbl>[,<fun1>,<val1>,...][,<arec>])" );
    elif IsOrdinaryTable( args[1] ) then
      tbls:= [ args[1] ];
    elif IsString( args[1] ) then
      tbls:= [ CharacterTable( args[1] ) ];
      if tbls[1] = fail then
        Error( "'", args[1],
               "' is not an admisible name of a character table" );
      fi;
    elif IsList( args[1] ) and ForAll( args[1], IsBlockOfCharacterTable ) then
      blocks:= args[1];
    elif IsList( args[1] ) and ForAll( args[1], IsOrdinaryTable ) then
      tbls:= args[1];
    elif IsList( args[1] ) and ForAll( args[1], IsString ) then
      tbls:= List( args[1], CharacterTable );
      if fail in tbls then
        Error( "not all entries in '", args[1],
               "' are admisible names of character tables" );
      fi;
    else
      Error( "<args>[1] must be either (a name of) an ord. char. table ",
             "or a list of (names of) ord. char. tables ",
             "or a list of block objects" );
    fi;

    if mode in [ "inv", "str" ] then
      # Determine the columns and the header.
      result:= rec( list:= [] );
      if IsEvenInt( Length( args ) ) then
        # The last argument must be a record that describes the columns
        # and perhaps the header.
        arec:= args[ Length( args ) ];
        if not IsRecord( arec ) then
          Error( "if <args> has even length its last entry must be a record" );
        elif IsBound( arec.columns ) then
          columns:= arec.columns;
        fi;
        if IsBound( arec.header ) then
          result.header:= arec.header;
        fi;
      fi;
      if IsBound( tbls ) and Length( tbls ) = 1 then
        # There is only one table.
        # Omit the 'G' column and add a header entry for the identifier
        # if columns and/or header are not explicitly given.
        if not IsBound( columns ) then
          columns:= Filtered( CTBlocks.BlockInvariantsColumns,
                              x -> x.label <> "G" );
        fi;
        if not IsBound( result.header ) then
          result.header:= [ Concatenation( "Block invariants for ",
                                           Identifier( tbls[1] ) ) ];
        fi;
      else
        # list of block objects or list of tables
        if not IsBound( columns ) then
          columns:= ShallowCopy( CTBlocks.BlockInvariantsColumns );
        fi;
        if not IsBound( result.header ) then
          result.header:= [ "Block invariants" ];
        fi;
      fi;

      replace:= function( col, B )
        if IsBound( col.replace ) then
          return col.replace[2][
                   Position( col.replace[1], col.fun( B ) ) ];
        else
          return String( col.fun( B ) );
        fi;
      end;
    elif mode in [ "one", "all" ] then
      result:= [];
    else
      Error( "<mode> must be one of \"all\", \"one\", \"inv\", \"str\"" );
    fi;

    # Scan the list of conditions.
    char:= fail;  # no restriction on p
    cond:= [];

    i:= 2;
    while i < Length( args ) do

      fun:= args[i];
      if not IsFunction( fun ) then
        Error( "<args>[", i, "] must be a function" );
      fi;
      vals:= args[ i+1 ];
      if not ( IsFunction( vals ) or IsList( vals ) ) then
        vals:= [ vals ];
      fi;

      # If the function restricts the characteristic
      # and if the list of candidate blocks is not prescribed
      # then use this information a priori, do not check it afterwards.
      if fun = UnderlyingCharacteristic and IsList( vals )
         and IsBound( tbls ) then
        char:= Filtered( vals, IsPrimeInt );
      else
        Add( cond, [ fun, vals ] );
      fi;

      # For the cases where invariants are computed:
      # If the function defines one of the result columns and if exactly
      # one value is prescribed then omit the column;
      # if no header was prescribed then also extend the header.
      if mode in [ "inv", "str" ] then
        if IsList( vals ) and Length( vals ) = 1 then
          r:= First( columns, x -> x.fun = fun );
          if fun <> fail then
            if not IsBound( arec ) or not IsBound( arec.columns ) then
              columns:= Filtered( columns, x -> x.label <> r.label );
            fi;
            if not IsBound( arec ) or not IsBound( arec.header ) then
              if IsBound( r.replace ) then
                pos:= Position( r.replace[1], vals[1] );
                if pos <> fail then
                  Add( result.header, r.replace[3][ pos ] );
                fi;
              else
                Add( result.header,
                     Concatenation( r.label, " = ", String( vals[1] ) ) );
              fi;
            fi;
          fi;
        fi;
      fi;

      i:= i + 2;
    od;

    # Create the list of relevant rows.
    if IsBound( tbls ) then
      for tbl in tbls do
        if char = fail then
          primes:= Set( Factors( Size( tbl ) ) );
          RemoveSet( primes, 1 );  # in case the group is trivial ...
        else
          primes:= char;
        fi;
        for p in primes do
          for B in PBlocks( tbl, p ) do
            if ForAll( cond, x ->
                 ( IsList( x[2] ) and x[1]( B ) in x[2] ) or
                 ( IsFunction( x[2] ) and x[2]( x[1]( B ) ) = true ) ) then
              if mode = "one" then
                return B;
              elif mode = "all" then
                Add( result, B );
              elif mode = "inv" then
                Add( result.list, List( columns, r -> r.fun( B ) ) );
              else
                Add( result.list, List( columns, r -> replace( r, B ) ) );
              fi;
            fi;
          od;
        od;
      od;
    else
      for B in blocks do
        if ForAll( cond, x ->
             ( IsList( x[2] ) and x[1]( B ) in x[2] ) or
             ( IsFunction( x[2] ) and x[2]( x[1]( B ) ) = true ) ) then
          if mode = "one" then
            return B;
          elif mode = "all" then
            Add( result, B );
          elif mode = "inv" then
            Add( result.list, List( columns, r -> r.fun( B ) ) );
          else
            Add( result.list, List( columns, r -> replace( r, B ) ) );
          fi;
        fi;
      od;
    fi;

    if mode = "one" then
      return fail;
    elif mode = "all" then
      return result;
    fi;

    result.columns:= columns;

    return result;
    end;


#############################################################################
##
#F  AllPBlocks( <tbl>[, <fun1>, <val1>, ...] )
##
InstallGlobalFunction( AllPBlocks, function( arg )
    return CTBlocks.AccessFunction( "AllPBlocks", arg, "all" );
    end );


#############################################################################
##
#F  OnePBlock( <tbl>[, <fun1>, <val1>, ...] )
##
InstallGlobalFunction( OnePBlock, function( arg )
    return CTBlocks.AccessFunction( "OnePBlock", arg, "one" );
    end );


#############################################################################
##
#F  BlockInvariants( <tbl>[, <fun1>, <val1>, ...][, <arec>] )
#F  BlockInvariants( <list>[, <arec>] )
##
InstallGlobalFunction( BlockInvariants, function( arg )
    return CTBlocks.AccessFunction( "BlockInvariants", arg, "inv" );
    end );


#############################################################################
##
#F  DisplayBlockInvariants( <tbls>[, <fun1>, <val1>, ...][, <arec>] )
#F  DisplayBlockInvariants( <list>[, <arec>] )
##
InstallGlobalFunction( DisplayBlockInvariants, function( arg )
    local onlyASCII, inv, posG, posp, entryG, entryp, line, result, width,
          toowide, i, ints, sum, j, pos, fun;

    onlyASCII:= CTblLib.ShowOnlyASCII();

    # Compute the data record.
    inv:= CTBlocks.AccessFunction( "DisplayBlockInvariants", arg, "str" );
    if Length( inv.list ) = 0 then
      return;
    fi;

    # Do not repeat the group name and the prime in each row.
    # (Assume that of both columns occur then "G" comes before "p".)
    posG:= PositionProperty( inv.columns, r -> r.label = "G" );
    posp:= PositionProperty( inv.columns, r -> r.label = "p" );
    if posG <> fail or posp <> fail then
      entryG:= "";
      entryp:= "";
      for line in inv.list do
        if posG <> fail then
          if line[ posG ] = entryG then
            line[ posG ]:= "";
          else
            entryG:= line[ posG ];
            entryp:= "";
          fi;
        fi;
        if posp <> fail then
          if line[ posp ] = entryp then
            line[ posp ]:= "";
          else
            entryp:= line[ posp ];
          fi;
        fi;
      od;
    fi;

    # Call 'StringStandardTable' for creating the list of output lines.
    result:= StringStandardTable( rec(
               header:= List( inv.columns, x -> x.label ),
               colalign:= List( inv.columns, x -> x.align ),
               matrix:= inv.list ) );

    if onlyASCII or not ( IsBound( TextAttr ) and IsRecord( TextAttr ) ) then
      result:= Concatenation( inv.header, [ "" ], result );
    else
      # Underline the fist header row.
      result:= Concatenation( [ Concatenation( [ TextAttr.underscore,
                                  inv.header[1], TextAttr.reset ] ) ],
                              inv.header{ [ 2 .. Length( inv.header ) ] },
                              [ "" ], result );
    fi;

    # Format the output.
    width:= SizeScreen()[1] - 2;
    result:= List( result,
               l -> InitialSubstringUTF8String( l, width, "*" ) );
    Add( result, "" );

    # Call the appropriate function.
    fun:= EvalString( UserPreference( "AtlasRep", "DisplayFunction" ) );
    fun( JoinStringsWithSeparator( result, "\n" ) );
    end );


#############################################################################
##
#F  BrowseBlockInvariants( <tbl>[, <fun1>, <val1>, ...][, <arec>] )
#F  BrowseBlockInvariants( <list>[, <arec>] )
##
##  We have to be aware of the possibility that the Browse package is not
##  loaded.
##
if not IsPackageMarkedForLoading( "Browse", "" ) then
  NCurses:= fail;
  BrowseData:= fail;
fi;

InstallGlobalFunction( BrowseBlockInvariants, function( arg )
    local inv, header, table, pos;

    if not IsBoundGlobal( "Browse" ) then
      return;
    fi;

    # Compute the data record.
    inv:= CTBlocks.AccessFunction( "DisplayBlockInvariants", arg, "str" );
    if Length( inv.list ) = 0 then
      return;
    fi;

    # Underline the first row of the header,
    # and add an empty line below the header.
    header:= Concatenation(
               [ [ NCurses.attrs.UNDERLINE, true,
                   inv.header[1],
                   NCurses.attrs.UNDERLINE, false ] ],
               inv.header{ [ 2 .. Length( inv.header ) ] },
               [ "" ] );

    # Construct the browse table.
    table:= rec(
      work:= rec(
        align:= "ct",
        header:= header,
        CategoryValues:= function( t, i, j )
          local col, val, repl;

          j:= j/2;
          col:= inv.columns[j];
          val:= t.work.main[ i/2 ][j];
          if IsBound( col.replace ) then
            repl:= col.replace;
            return [ BrowseData.ReplacedEntry( val, repl[2], repl[3] ) ];
          else
            return [ Concatenation( col.label, " = ", val ) ];
          fi;
        end,

        main:= inv.list,
        labelsRow:= [],
        labelsCol:= [ List( inv.columns,
                            r -> rec( rows:= [ r.label ],
                                      align:= [ r.align[1] ] ) ) ],
        sepLabelsCol:= "=",
        sepRow:= "-",
        sepCol:= Concatenation( [ "| " ],
            ListWithIdenticalEntries( Length( inv.columns ) - 1, " | " ),
            [ " |" ] ),

        SpecialGrid:= BrowseData.SpecialGridLineDraw,
      ),
      dynamic:= rec(
        sortFunctionsForColumns:= ListWithIdenticalEntries(
          Length( inv.columns ), BrowseData.CompareLenLex ),
      ),
    );

    pos:= PositionProperty( inv.columns, r -> r.label = "G" );
    if pos <> fail then
      table.dynamic.sortFunctionsForColumns[ pos ]:=
        BrowseData.defaults.dynamic.sortFunctionForColumnsDefault;
    fi;

    NCurses.BrowseGeneric( table );
    end );

if not IsPackageMarkedForLoading( "Browse", "" ) then
  Unbind( NCurses );
  Unbind( BrowseData );
fi;


##############################################################################
##
#F  SCAlgebraCentreOfGroupAlgebra( <G>, <p> )
#F  SCAlgebraCentreOfGroupAlgebra( <G>, <F> )
#F  SCAlgebraCentreOfGroupAlgebra( <tbl>, <p> )
#F  SCAlgebraCentreOfGroupAlgebra( <tbl>, <F> )
##
InstallGlobalFunction( SCAlgebraCentreOfGroupAlgebra,
    function( G_or_tbl, p_or_F )
    local F, zero, one, nccl, T, classes, reps, i, j, list, c, g, h, k,
          llist, onecoeffs, A;

    if IsInt( p_or_F ) and IsPrimeInt( p_or_F ) then
      F:= GF( p_or_F );
    elif IsField( p_or_F ) and IsFinite( p_or_F ) then
      F:= p_or_F;
    else
      Error( "<p_or_F> must be a prime integer or a finite field" );
    fi;

    zero:= Zero( F );
    one:= One( F );
    nccl:= NrConjugacyClasses( G_or_tbl );
    T:= EmptySCTable( nccl, zero, "symmetric" );

    if IsGroup( G_or_tbl ) then
      # This is a very simpleminded implementation.
      classes:= ConjugacyClasses( G_or_tbl );
      reps:= List( classes, Representative );
      for i in [ 1 .. nccl ] do
        for j in [ 1 .. i ] do
          list:= 0 * [ 1 .. nccl ];
          c:= 0;
          for g in classes[i] do
            for h in classes[j] do
              k:= Position( reps, g * h );
              if k <> fail then
                list[k]:= list[k] + 1;
              fi;
            od;
          od;
          llist:= [];
          for k in [ 1 .. nccl ] do
            c:= list[k];
            if c <> 0 then
              Append( llist, [ c * one, k ] );
            fi;
          od;
          SetEntrySCTable( T, i, j, llist );
        od;
      od;
    elif IsOrdinaryTable( G_or_tbl ) then
      for i in [ 1 .. nccl ] do
        for j in [ 1 .. i ] do
          list:= [];
          for k in [ 1 .. nccl ] do
            c:= ClassMultiplicationCoefficient( G_or_tbl, i, j, k );
            if c <> 0 then
              Append( list, [ c * one, k ] );
            fi;
          od;
          SetEntrySCTable( T, i, j, list );
        od;
      od;
    else
      Error( "<G_or_tbl> must be a group or an ord. char. table" );
    fi;

    onecoeffs:= zero * [ 1 .. nccl ];
    onecoeffs[1]:= one;

    A:= AlgebraWithOneByStructureConstants( F, T,
            List( [ 1 .. nccl ],
                  i -> Concatenation( "C", String( i ) ) ),
            onecoeffs );
    SetIsAssociative( A, true );
    SetIsCommutative( A, true );

#T an ugly hack ...
#T needed for what?
if IsGroup( G_or_tbl ) then
  A!.classlengths:= List( ConjugacyClasses( G_or_tbl ), Size );
else
  A!.classlengths:= SizesConjugacyClasses( G_or_tbl );
fi;
    return A;
    end );
#T see better.g: replace the code?


##############################################################################
##
#F  BlockDecompositionSCAlgebraCentreOfGroupAlgebra( <tbl>, <p> )
#F  BlockDecompositionSCAlgebraCentreOfGroupAlgebra( <tbl>, <F> )
##
InstallGlobalFunction( BlockDecompositionSCAlgebraCentreOfGroupAlgebra,
    function( tbl, p_or_F )
    local p, bl, coeffs, concat, F, A, fam, T, gens, prodslist;

    if IsInt( p_or_F ) and IsPrimeInt( p_or_F ) then
      p:= p_or_F;
    elif IsField( p_or_F ) and IsFinite( p_or_F ) then
      p:= Characteristic( p_or_F );
    else
      Error( "<p_or_F> must be a prime integer or a finite field" );
    fi;

    bl:= PrimeBlocks( tbl, p );

    coeffs:= List( [ 1 .. Length( bl.defect ) ],
                   b -> List( CoefficientsOfOsimaIdempotent( tbl, p, b ),
                              x -> FrobeniusCharacterValue( x, p ) ) );
    concat:= Concatenation( coeffs );

    if IsInt( p_or_F ) then
      F:= Field( concat );
    elif IsSubset( p_or_F, concat ) then
      F:= p_or_F;
    else
      return fail;
    fi;

    A:= SCAlgebraCentreOfGroupAlgebra( tbl, F );
    fam:= ElementsFamily( FamilyObj( A ) );
    T:= fam!.sctable;
    gens:= List( GeneratorsOfAlgebra( A ), ExtRepOfObj );
    prodslist:= List( coeffs,
                      e -> List( gens, x -> SCTableProduct( T, x, e ) ) );

    return List( prodslist,
                 prods -> SubalgebraNC( A,
                     List( SemiEchelonMatDestructive( prods ).vectors,
                           x -> ObjByExtRep( fam, x ) ), "basis" ) );
    end );


##############################################################################
##
#F  SCAlgebraCentreOfBlock( <tbl>, <p>, <b> )
#F  SCAlgebraCentreOfBlock( <tbl>, <F>, <b> )
##
InstallGlobalFunction( SCAlgebraCentreOfBlock, function( tbl, p_or_F, b )
    local p, coeffs, F, A, fam, T, prods;

    if IsInt( p_or_F ) and IsPrimeInt( p_or_F ) then
      p:= p_or_F;
    elif IsField( p_or_F ) and IsFinite( p_or_F ) then
      p:= Characteristic( p_or_F );
    else
      Error( "<p_or_F> must be a prime integer or a finite field" );
    fi;

    coeffs:= CoefficientsOfOsimaIdempotent( tbl, p, b );
    if coeffs = fail then
      return fail;
    fi;
    coeffs:= List( coeffs, x -> FrobeniusCharacterValue( x, p ) );

    if IsInt( p_or_F ) then
      F:= Field( coeffs );
    elif IsSubset( p_or_F, coeffs ) then
      F:= p_or_F;
    else
      return fail;
    fi;

    A:= SCAlgebraCentreOfGroupAlgebra( tbl, F );
    fam:= ElementsFamily( FamilyObj( A ) );
    T:= fam!.sctable;
    prods:= List( GeneratorsOfAlgebra( A ),
                  x -> SCTableProduct( T, ExtRepOfObj( x ), coeffs ) );

    return SubalgebraNC( A, List( SemiEchelonMatDestructive( prods ).vectors,
                                  x -> ObjByExtRep( fam, x ) ), "basis" );
    end );


#############################################################################
##
#E

