#############################################################################
##
#W  utils.g              GAP 4 package CTBlocks                 Thomas Breuer
##
#Y  Copyright (C)  2013,   Lehrstuhl D für Mathematik,  RWTH Aachen,  Germany
##
##  This file contains some utility functions for the CTBlocks package.
##


if not IsBound( GaloisData ) then

#############################################################################
##
#A  GaloisData( <tbl> )
##
##  For a character table <A>tbl</A> with <M>n</M> conjugacy classes,
##  this function returns a record with the following components.
##  <P/>
##  <List>
##  <Mark>proxy</Mark>
##  <Item>
##    a list of <M>n</M> positive integers such that the <M>i</M>-th entry
##    is the position of the first class of <A>tbl</A> that is Galois
##    conjugate to the <M>i</M>-th class,
##  </Item>
##  <Mark>power</Mark>
##  <Item>
##    a list of <M>n</M> integers such that the elements in the <M>i</M>-th
##    class of <A>tbl</A> are obtained by taking the
##    <C>power</C><M>[i]</M>-th powers of the elements in the
##    <C>proxy</C><M>[i]</M>-th class.
##  </Item>
##  </List>
##
DeclareAttributeSuppCT( "GaloisData", IsCharacterTable, [ "class" ] );


#############################################################################
##
#M  GaloisData( <tbl> )
##
InstallMethod( GaloisData,
    [ "IsCharacterTable" ],
    function( tbl )
    local galois, proxy, power, i, entry, j;

    galois:= GaloisMat( TransposedMat( Irr( tbl ) ) ).galoisfams;
    proxy:= [];
    power:= [];
    for i in [ 1 .. Length( galois ) ] do
      entry:= galois[i];
      if entry = 1 then
        proxy[i]:= i;
        power[i]:= 1;
      elif IsList( entry ) then
        for j in [ 1 .. Length( entry[1] ) ] do
          proxy[ entry[1][j] ]:= i;
          power[ entry[1][j] ]:= entry[2][j];
        od;
      fi;
    od;

    return rec( proxy:= proxy, power:= power );
    end );

fi;


#############################################################################
##
#F  PrintOverviewOfDefectOneNormalizers( <tbl> )
##
##  <#GAPDoc Label="PrintOverviewOfDefectOneNormalizers">
##  <ManSection>
##  <Func Name="PrintOverviewOfDefectOneNormalizers" Arg="tbl"/>
##
##  <Description>
##  Let <A>tbl</A> be the ordinary character table of a finite group
##  <M>G</M>, say.
##  For each prime integer <M>p</M> that divides the order of <M>G</M>
##  exactly once, <Ref Func="PrintOverviewOfDefectOneNormalizers"/> prints
##  one or two lines of information about the normalizers of
##  Sylow <M>p</M>-subgroups in <M>G</M>.
##  <P/>
##  If such a Sylow <M>p</M>-normalizer is maximal in <M>G</M> and if the
##  attribute <Ref Attr="Maxes" BookName="CTblLib"/> is set in <A>tbl</A>
##  then one line is printed for <M>p</M>, saying that the normalizer is
##  maximal, and showing the name of the character table.
##  <P/>
##  Otherwise, if the character table of the Sylow <M>p</M>-normalizer is
##  known then the availability of the character table and its name are
##  mentioned, and if not then an approximation of the structure is shown.
##  If the attribute <Ref Attr="Maxes" BookName="CTblLib"/> is set in
##  <A>tbl</A> then also the list of maximal subgroups is shown that contain
##  a Sylow <M>p</M>-normalizer.
##  <P/>
##  <Example><![CDATA[
##  gap> PrintOverviewOfDefectOneNormalizers( CharacterTable( "A5" ) );
##  3: max. subgroup S3
##  5: max. subgroup D10
##  gap> PrintOverviewOfDefectOneNormalizers( CharacterTable( "A6" ) );
##  5: char. table available as D10,
##     contained in max. subgroups A5, A6M2
##  gap> PrintOverviewOfDefectOneNormalizers( CharacterTable( "S10" ) );
##  7: structure (7x[6]).6,
##     contained in max. subgroups S7xS3
##  ]]></Example>
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
##
BindGlobal( "PrintOverviewOfDefectOneNormalizers", function( tbl )
    local facts, len, space, orders, maxes, pair, p, cls, cen, norm, maxinfo,
          super, mx, subcls, subcen, sub;

    facts:= Filtered( Collected( Factors( Size( tbl ) ) ), x -> x[2] = 1 );
    if facts[1][1] = 1 then
      # The group is trivial.
      return;
    fi;
    len:= LogInt( facts[ Length( facts ) ][1], 10 ) + 1;
    space:= FormattedString( "", len + 2 );
    orders:= OrdersClassRepresentatives( tbl );
    if HasMaxes( tbl ) then
      maxes:= List( Maxes( tbl ), CharacterTable );
    else
      maxes:= fail;
    fi;

    for pair in facts do
      p:= pair[1];
      cls:= Positions( orders, p );
      cen:= SizesCentralizers( tbl )[ cls[1] ];
      norm:= Phi( p ) / Length( cls );
      maxinfo:= false;
      if maxes <> fail then
        super:= [];
        for mx in maxes do
          subcls:= Positions( OrdersClassRepresentatives( mx ), p );
          if not IsEmpty( subcls ) then
            subcen:= SizesCentralizers( mx )[ subcls[1] ];
            if subcen = cen and Length( subcls ) = Length( cls ) then
              Add( super, mx );
            fi;
          fi;
        od;

        if Length( super ) = 1 and Size( super[1] ) = cen * norm then
          maxinfo:= true;
        else
          maxinfo:= List( super, Identifier );
        fi;
      fi;

      Print( FormattedString( p, len ), ": " );

      if maxinfo = true then
        Print( "max. subgroup ", Identifier( super[1] ) );
      else
        sub:= CharacterTable( Concatenation( Identifier( tbl ), "N",
                                             String( p ) ) );
        if sub <> fail then
          Print( "char. table available as ", Identifier( sub ) );
        elif cen = p then
          Print( "structure ", String( p ), ":", String( norm ) );
        elif IsPrimeInt( cen / p ) then
          Print( "structure (", String( p ), "x", String( cen / p ), ").",
                 String( norm ) );
        else
          Print( "structure (", String( p ), "x[", String( cen / p ), "]).",
                 String( norm ) );
        fi;
        if maxinfo <> false then
          Print( ",\n", space, "contained in max. subgroups ",
                 JoinStringsWithSeparator( maxinfo, ", " ) );
        fi;
      fi;

      Print( "\n" );
    od;
end );


#############################################################################
##
#F  AlgebraWithOneByStructureConstants( <R>, <sctable>, <onecoeffs> )
#F  AlgebraWithOneByStructureConstants( <R>, <sctable>, <name>, <onecoeffs> )
#F  AlgebraWithOneByStructureConstants( <R>, <sctable>, <names>, <onecoeffs> )
#F  AlgebraWithOneByStructureConstants( <R>, <sctable>, <name1>, <name2>,
#F                                      ..., <onecoeffs> )
##
##  Do not document this function, it will eventually become part of the
##  GAP library.
##
##  <ManSection>
##  <Func Name="AlgebraWithOneByStructureConstants"
##   Arg="R, sctable[, nameinfo], onecoeffs"/>
##
##  <Description>
##  The only differences between this function and
##  <Ref Func="AlgebraByStructureConstants" BookName="ref"/> are that
##  <Ref Func="AlgebraWithOneByStructureConstants"/> takes an additional
##  argument <A>onecoeffs</A>, the list over the ring <A>R</A> that describes
##  the unique multiplicative identity element of the algebra
##  that is returned, w. r. t. the defining basis of this algebra,
##  and that the returned algebra is an algebra-with-one
##  (see <Ref Func="IsAlgebraWithOne" BookName="ref"/>).
##  <P/>
##  This way, one avoids computations that have to establish the identity
##  element.
##  <P/>
##  <Example><![CDATA[
##  gap> A:= GF(2)^[2,2];;
##  gap> B:= Basis( A );;
##  gap> onecoeffs:= Coefficients( B, One( A ) );
##  [ Z(2)^0, 0*Z(2), 0*Z(2), Z(2)^0 ]
##  gap> T:= StructureConstantsTable( B );;
##  gap> One( a );
##  gap> sc1:= AlgebraByStructureConstants( GF(2), T );
##  <algebra of dimension 4 over GF(2)>
##  gap> HasOne( sc1 );
##  false
##  gap> One( sc1 );
##  v.1+v.4
##  gap> sc2:= AlgebraWithOneByStructureConstants( GF(2), T, onecoeffs );
##  <algebra-with-one of dimension 4 over GF(2)>
##  gap> HasOne( sc2 );
##  true
##  gap> One( sc2 );
##  v.1+v.4
##  ]]></Example>
##  </Description>
##  </ManSection>
##

if not IsBound( AlgebraWithOneByStructureConstants ) then

BindGlobal( "AlgebraWithOneByStructureConstants", function( arg )
    local A, Fam, one;

    A:= CallFuncList( AlgebraByStructureConstants,
                      arg{ [ 1 .. Length( arg )-1 ] } );
    Fam:= ElementsFamily( FamilyObj( A ) );
    one:= ObjByExtRep( Fam, arg[ Length( arg ) ] );

    SetOne( Fam, one );
    SetOne( A, one );
    SetFilterObj( A, IsMagmaWithOne );

    return A;
    end );

fi;


############################################################################
##
#F  CTBlocks.SimplifiedStraightLineProgram( <slp>[, <gens>] )
##
##  This is an experimental function.
##  Its aim is to compute a simplified version of the straight line program
##  <slp>.
##  If <gens> is given then more criteria can be applied,
##  and the result is guaranteed to compute the same result on <gens> as
##  <slp> does.
##
##  The function is useful in particular if <slp> has been computed by
##  'FindShortGeneratorsOfSubgroup'.
##
##  The criteria are as follows:
##  - If the same group element arises in different positions
##    then keep only the first (use only references to this one).
##  - Omit identity elements.
##  - Omit powerings to the first power.
##  - Omit inversions of involutions.
##  - Replace inverting by squaring for elements of order three.
##
CTBlocks.SimplifiedStraightLineProgram:= function( slp, gens... )
    local lines, len, nrinputs, orders, intermed, one, nextfree,
          i, line, j, pos, lline, k, used, mapping, offset, range;

    lines:= List( LinesOfStraightLineProgram( slp ), ShallowCopy );
    if not ForAll( lines, line -> ForAll( line, IsInt ) ) then
      Error( "sorry, the general case is not yet supported" );
    fi;
    len:= Length( lines );
    nrinputs:= NrInputsOfStraightLineProgram( slp );

    if Length( gens ) = 1 then
      gens:= gens[1];
      orders:= List( gens, Order );
      intermed:= ShallowCopy( gens );
      one:= One( gens[1] );
    fi;

    nextfree:= nrinputs + 1;
    for i in [ 1 .. len ] do
      line:= lines[i];
      if ForAll( line, IsInt ) then

        # If generators are known then check whether the group element
        # computed with 'line' occurs earlier;
        # in this case, replace the computation by a cross-reference.
        if IsBound( gens ) then
          intermed[i+nrinputs]:= one;
          for j in [ 1, 3 .. Length( line )-1 ] do
            intermed[ i + nrinputs ]:= intermed[ i + nrinputs ]
                * intermed[ line[j] ]^line[ j+1 ];
          od;
          orders[ i + nrinputs ]:= Order( intermed[ i + nrinputs ] );
          pos:= Position( intermed, intermed[ i + nrinputs ] );
          if pos < i+nrinputs then
            lines[i]:= [ pos, 1 ];
            line:= lines[i];
          fi;
        fi;

        if   Length( line ) = 2 and line[2] = 0 then

          # If the result of a line is an identity element then
          # do not use it in later computations.
          for j in [ i+1 .. len ] do
            lline:= lines[j];
            for k in [ 1, 3 .. Length( lline )-1 ] do
              if lline[k] = nextfree then
                Unbind( lline[k] );
                Unbind( lline[ k+1 ] );
              fi;
            od;
            if IsEmpty( lline ) then
              # The same code will remove this line later.
              lines[j]:= [ 1, 0 ];
            else
              lines[j]:= Compacted( lline );
            fi;
          od;

          # Remove the unnecessary line.
          Unbind( lines[i] );

        elif Length( line ) = 2 and
             ( line[2] = 1 or ( line[2] = -1 and
               IsBound( orders ) and orders[ line[1] ] = 2 ) ) then

          # Avoid raising elements to the first power,
          # and inverting involutions;
          # replace the result by the original element in later lines.
          for j in [ i+1 .. len ] do
            lline:= lines[j];
            for k in [ 1, 3 .. Length( lline )-1 ] do
              if lline[k] = nextfree then
                lline[k]:= line[1];
              fi;
            od;
          od;

          # Remove the unnecessary line.
          Unbind( lines[i] );

        elif Length( line ) = 2 and line[2] = -1 and
             IsBound( orders ) and orders[ line[1] ] = 3 then
          # Replace inverting by squaring for elements of order three.
          lines[i]:= [ line[1], 2 ];
        fi;

        nextfree:= nextfree + 1;

      else
        Error( "not available for <line> not consisting of integers" );
      fi;
    od;

    # Remove all lines whose result is not used at all,
    # except for the result line.
    used:= List( [ 2 .. Length( lines ) + nrinputs ], x -> 0 );
    for line in lines do
      for i in [ 1, 3 .. Length( line ) - 1 ] do
        j:= line[i];
        used[j]:= used[j] + 1;
      od;
    od;
    for i in Filtered( Positions( used, 0 ) - nrinputs, IsPosInt ) do
      Unbind( lines[i] );
    od;

    # Remove the holes which we have created.
    nextfree:= nrinputs + 1;
    mapping:= [ 1 .. nrinputs ];
    offset:= 0;
    for i in [ 1 .. len ] do
      if IsBound( lines[i] ) then
        line:= lines[i];
        if ForAll( line, IsInt ) then
          range:= [ 1, 3 .. Length( line )-1 ];
          line{ range }:= mapping{ line{ range } };
          mapping[ i + nrinputs ]:= nextfree;
          nextfree:= nextfree + 1;
        else
          Error( "not available for <line> not consisting of integers" );
        fi;
      else
        offset:= offset + 1;
        mapping[ i + nrinputs ]:= fail;
      fi;
    od;

    return StraightLineProgram( MakeImmutable( Compacted( lines ) ),
                                nrinputs );
    end;


#############################################################################
##
#F  CTBlocks.SLPsForSubgroupsOfLowIndex( <G>, <n> )
##
##  This is an experimental function.
##  Its aim is to compute straight line programs for representatives of
##  conjugacy classes of subgroups of index exactly <n> in the group <G>,
##  using a presentation of <G> (via 'IsomorphismFpGroupByGeneratorsNC')
##  and 'LowIndexSubgroupsFpGroup'.
##
##  Note that 'FindShortGeneratorsOfSubgroup' can yield better results.
##
CTBlocks.SLPsForSubgroupsOfLowIndex:= function( G, n )
    local gens, gensorders, iso, img, low, result, freegens, sub, subgens,
          Gsubgens, Gsub, init, len, size, i, test, gen;

    gens:= GeneratorsOfGroup( G );
    gensorders:= List( gens, Order );
    iso:= IsomorphismFpGroupByGeneratorsNC( G, gens, "x" );
    img:= Images( iso );
    low:= LowIndexSubgroupsFpGroup( img, n );

    result:= [];
    freegens:= FreeGeneratorsOfFpGroup( img );
    for sub in Filtered( low, x -> Index( img, x ) = n ) do
      subgens:= List( GeneratorsOfGroup( sub ), x -> x![1] );
      Gsubgens:= List( subgens, x -> MappedWord( x, freegens, gens ) );
      Gsub:= SubgroupNC( G, Gsubgens );

      # Try to omit some generators.
      init:= [];
      len:= Length( subgens );
      size:= Size( Gsub );
      for i in [ 1 .. len ] do
        test:= SubgroupNC( Gsub, Concatenation( Gsubgens{ init },
                                     Gsubgens{ [ i+1 .. len ] } ) );
        if Size( test ) < size then
          Add( init, i );
        fi;
      od;
      subgens:= subgens{ init };
      Gsubgens:= Gsubgens{ init };

      # Avoid unnecessary negative powers.
      subgens:= List( subgens, x -> ShallowCopy( ExtRepOfObj( x ) ) );
      for gen in subgens do
        for i in [ 1, 3 .. Length( gen )-1 ] do
          if not ( gen[ i+1 ] = -1 and gensorders[ gen[i] ] > 3 ) then
            gen[ i+1 ]:= gen[ i+1 ] mod gensorders[ gen[i] ];
          fi;
        od;
      od;

      Add( result, [ subgens, SubgroupNC( Gsub, Gsubgens ) ] );
    od;

    return result;
    end;


#############################################################################
##
#E

